import re

from loguru import logger

from flexget import plugin
from flexget.components.imdb.utils import extract_id, is_valid_imdb_title_id
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import json
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException
from flexget.utils.soup import get_soup

logger = logger.bind(name='imdb_watchlist')
USER_ID_RE = r'^ur\d{7,9}$'
CUSTOM_LIST_RE = r'^ls\d{7,10}$'
USER_LISTS = ['watchlist', 'ratings', 'checkins']
TITLE_TYPE_MAP = {
    'movies': 'movie',
    'short films': 'short',
    'games': 'videoGame',
    'mini series': 'tvMiniSeries',
    'shows': 'tvSeries',
    'episodes': 'tvEpisode',
    'tv movies': 'tvMovie',
    'tv specials': 'tvSpecial',
    'videos': 'video',
}


class ImdbWatchlist:
    """"Creates an entry for each movie in your imdb list."""

    schema = {
        'type': 'object',
        'properties': {
            'user_id': {
                'type': 'string',
                'pattern': USER_ID_RE,
                'error_pattern': 'user_id must be in the form urXXXXXXX',
            },
            'list': {
                'type': 'string',
                'oneOf': [{'enum': USER_LISTS}, {'pattern': CUSTOM_LIST_RE}],
                'error_oneOf': 'list must be either %s, or a custom list name (lsXXXXXXXXX)'
                % ', '.join(USER_LISTS),
            },
            'force_language': {'type': 'string', 'default': 'en-us'},
            'type': {
                'oneOf': [
                    one_or_more(
                        {'type': 'string', 'enum': list(TITLE_TYPE_MAP.keys())}, unique_items=True
                    ),
                    {'type': 'string', 'enum': ['all']},
                ]
            },
        },
        'additionalProperties': False,
        'required': ['list'],
        'anyOf': [
            {'required': ['user_id']},
            {'properties': {'list': {'pattern': CUSTOM_LIST_RE}}},
        ],
        'error_anyOf': 'user_id is required if not using a custom list (lsXXXXXXXXX format)',
    }

    def prepare_config(self, config):
        if 'type' not in config:
            config['type'] = ['all']

        if not isinstance(config['type'], list):
            config['type'] = [config['type']]

        return config

    @cached('imdb_watchlist', persist='2 hours')
    def on_task_input(self, task, config):
        config = self.prepare_config(config)

        # Create movie entries by parsing imdb list page(s) html using beautifulsoup
        logger.verbose('Retrieving imdb list: {}', config['list'])

        headers = {'Accept-Language': config.get('force_language')}
        params = {'view': 'detail', 'page': 1}
        if config['list'] in ['watchlist', 'ratings', 'checkins']:
            url = 'http://www.imdb.com/user/%s/%s' % (config['user_id'], config['list'])
            if config['list'] == 'watchlist':
                params = {'view': 'detail'}
        else:
            url = 'http://www.imdb.com/list/%s' % config['list']
        if 'all' not in config['type']:
            title_types = [TITLE_TYPE_MAP[title_type] for title_type in config['type']]
            params['title_type'] = ','.join(title_types)
            params['sort'] = 'list_order%2Casc'

        if config['list'] == 'watchlist':
            entries = self.parse_react_widget(task, config, url, params, headers)
        else:
            entries = self.parse_html_list(task, config, url, params, headers)
        return entries

    def fetch_page(self, task, url, params, headers):
        logger.debug('Requesting: {} {}', url, headers)
        try:
            page = task.requests.get(url, params=params, headers=headers)
        except RequestException as e:
            raise plugin.PluginError(str(e))
        if page.status_code != 200:
            raise plugin.PluginError(
                'Unable to get imdb list. Either list is private or does not exist.'
                + ' Html status code was: %d.' % page.status_code
            )
        return page

    def parse_react_widget(self, task, config, url, params, headers):
        page = self.fetch_page(task, url, params, headers)
        try:
            json_vars = json.loads(
                re.search(r'IMDbReactInitialState.push\((.+?)\);\n', page.text).group(1)
            )
        except (TypeError, AttributeError, ValueError) as e:
            raise plugin.PluginError(
                'Unable to get imdb list from imdb react widget.'
                + ' Either the list is empty or the imdb parser of the imdb_watchlist plugin is broken.'
                + ' Original error: %s.' % str(e)
            )
        total_item_count = 0
        if 'list' in json_vars and 'items' in json_vars['list']:
            total_item_count = len(json_vars['list']['items'])
        if not total_item_count:
            logger.verbose('No movies were found in imdb list: {}', config['list'])
            return []
        imdb_ids = []
        for item in json_vars['list']['items']:
            if is_valid_imdb_title_id(item.get('const')):
                imdb_ids.append(item['const'])
        params = {'ids': ','.join(imdb_ids)}
        url = 'http://www.imdb.com/title/data'
        try:
            json_data = self.fetch_page(task, url, params, headers).json()
        except (ValueError, TypeError) as e:
            raise plugin.PluginError(
                'Unable to get imdb list from imdb JSON API.'
                + ' Either the list is empty or the imdb parser of the imdb_watchlist plugin is broken.'
                + ' Original error: %s.' % str(e)
            )
        logger.verbose('imdb list contains {} items', len(json_data))
        logger.debug(
            'First entry (imdb id: {}) looks like this: {}', imdb_ids[0], json_data[imdb_ids[0]]
        )
        entries = []
        for imdb_id in imdb_ids:
            entry = Entry()
            if not (
                'title' in json_data[imdb_id]
                and 'primary' in json_data[imdb_id]['title']
                and 'href' in json_data[imdb_id]['title']['primary']
                and 'title' in json_data[imdb_id]['title']['primary']
            ):
                logger.debug('no title or link found for item {}, skipping', imdb_id)
                continue
            if 'type' in json_data[imdb_id]['title']:
                entry['imdb_type'] = json_data[imdb_id]['title']['type']
            title = json_data[imdb_id]['title']['primary']['title']
            entry['title'] = title
            if 'year' in json_data[imdb_id]['title']['primary']:
                year = json_data[imdb_id]['title']['primary']['year'][0]
                entry['title'] += ' (%s)' % year
                entry['imdb_year'] = year
            entry['url'] = json_data[imdb_id]['title']['primary']['href']
            entry['imdb_id'] = imdb_id
            entry['imdb_name'] = entry['title']
            entries.append(entry)
        return entries

    def parse_html_list(self, task, config, url, params, headers):
        page = self.fetch_page(task, url, params, headers)
        soup = get_soup(page.text)
        try:
            item_text = soup.find('div', class_='lister-total-num-results').string.split()
            total_item_count = int(item_text[0].replace(',', ''))
            logger.verbose('imdb list contains {} items', total_item_count)
        except AttributeError:
            total_item_count = 0
        except (ValueError, TypeError) as e:
            # TODO Something is wrong if we get a ValueError, I think
            raise plugin.PluginError(
                'Received invalid movie count: %s ; %s'
                % (soup.find('div', class_='lister-total-num-results').string, e)
            )

        if not total_item_count:
            logger.verbose('No movies were found in imdb list: {}', config['list'])
            return

        entries = []
        items_processed = 0
        page_no = 1
        while items_processed < total_item_count:
            # Fetch the next page unless we've just begun
            if items_processed:
                page_no += 1
                params['page'] = page_no
                page = self.fetch_page(task, url, params, headers)
                soup = get_soup(page.text)

            items = soup.find_all('div', class_='lister-item')
            if not items:
                logger.debug('no items found on page: {}, aborting.', url)
                break
            logger.debug('{} items found on page {}', len(items), page_no)

            for item in items:
                items_processed += 1
                a = item.find('h3', class_='lister-item-header').find('a')
                if not a:
                    logger.debug('no title link found for row, skipping')
                    continue

                link = ('http://www.imdb.com' + a.get('href')).rstrip('/')
                entry = Entry()
                entry['title'] = a.text
                try:
                    year = int(item.find('span', class_='lister-item-year').text)
                    entry['title'] += ' (%s)' % year
                    entry['imdb_year'] = year
                except (ValueError, TypeError):
                    pass
                entry['url'] = link
                entry['imdb_id'] = extract_id(link)
                entry['imdb_name'] = entry['title']
                entries.append(entry)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(ImdbWatchlist, 'imdb_watchlist', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.config_schema import process_config
from flexget.event import event
from flexget.plugin import PluginError

from . import series as plugin_series

logger = logger.bind(name='configure_series')


class ConfigureSeries(plugin_series.FilterSeriesBase):
    """Generates series configuration from any input (supporting API version 2, soon all)

    Configuration::

      configure_series:
        [settings]:
           # same configuration as series plugin
        from:
          [input plugin]: <configuration>

    Example::

      configure_series:
        settings:
          quality: 720p
        from:
          listdir:
            - /media/series
    """

    @property
    def schema(self):
        return {
            'type': 'object',
            'properties': {
                'settings': self.settings_schema,
                'from': {'$ref': '/schema/plugins?phase=input'},
            },
            'required': ['from'],
            'additionalProperties': False,
        }

    def on_task_prepare(self, task, config):

        series = {}
        for input_name, input_config in config.get('from', {}).items():
            input_plugin = plugin.get_plugin_by_name(input_name)
            method = input_plugin.phase_handlers['input']
            try:
                result = method(task, input_config)
            except PluginError as e:
                logger.warning('Error during input plugin {}: {}', input_name, e)
                continue
            if not result:
                logger.warning('Input {} did not return anything', input_name)
                continue

            for entry in result:
                s = series.setdefault(entry['title'], {})
                if entry.get('tvdb_id'):
                    s['set'] = {'tvdb_id': entry['tvdb_id']}

                # Allow configure_series to set anything available to series
                for key, schema in self.settings_schema['properties'].items():
                    if 'configure_series_' + key in entry:
                        errors = process_config(
                            entry['configure_series_' + key], schema, set_defaults=False
                        )
                        if errors:
                            logger.debug(
                                'not setting series option {} for {}. errors: {}',
                                key,
                                entry['title'],
                                errors,
                            )
                        else:
                            s[key] = entry['configure_series_' + key]

        if not series:
            logger.info('Did not get any series to generate series configuration')
            return

        # Make a series config with the found series
        # Turn our dict of series with settings into a list of one item dicts
        series_config = {'generated_series': [dict([x]) for x in series.items()]}
        # If options were specified, add them to the series config
        if 'settings' in config:
            series_config['settings'] = {'generated_series': config['settings']}
        # Merge our series config in with the base series config
        self.merge_config(task, series_config)


@event('plugin.register')
def register_plugin():
    plugin.register(ConfigureSeries, 'configure_series', api_ver=2)



import logging
import webbrowser
from functools import partial, wraps
from pathlib import Path
from typing import List, Optional, Callable

from loguru import logger

from flexget import __version__

logger = logger.bind(name='tray_icon')

try:
    # If we are running outside of a graphical environment, these imports will fail
    from PIL import Image
    from pystray import Icon, Menu, MenuItem

    _import_success = True
except Exception as e:
    logger.debug('Could not load tray icon: {}', e)
    _import_success = False


def check_if_tray_is_active(f):
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        if not self.active:
            return
        return f(self, *args, **kwargs)

    return wrapped


image_path = Path(__file__).parent / 'resources' / 'flexget.png'


class TrayIcon:
    def __init__(self, path_to_image: Path = image_path):
        # Silence PIL noisy logging
        logging.getLogger('PIL.PngImagePlugin').setLevel(logging.INFO)
        logging.getLogger('PIL.Image').setLevel(logging.INFO)

        self.path_to_image: Path = path_to_image
        self.icon: Optional['Icon'] = None
        self._menu: Optional['Menu'] = None
        self.menu_items: List['MenuItem'] = []

        self.active: bool = _import_success
        self.running: bool = False

        self.add_core_menu_items()

    @check_if_tray_is_active
    def add_menu_item(
        self,
        text: str = None,
        action: Callable = None,
        menu_item: 'MenuItem' = None,
        index: int = None,
        **kwargs,
    ):
        """
        Add a menu item byt passing its text and function, or pass a created MenuItem. Force position by sending index
        """
        if not any(v for v in (menu_item, text)):
            raise ValueError(f"Either 'text' or 'menu_item' are required")

        menu_item = menu_item or MenuItem(text=text, action=action, **kwargs)
        if index is not None:
            self.menu_items.insert(index, menu_item)
        else:
            self.menu_items.append(menu_item)

    @check_if_tray_is_active
    def add_menu_separator(self, index: int = None):
        self.add_menu_item(menu_item=Menu.SEPARATOR, index=index)

    def add_core_menu_items(self):
        open_web = partial(webbrowser.open)
        self.add_menu_item(text=f'Flexget {__version__}', enabled=False)
        self.add_menu_separator()
        self.add_menu_item(text='Homepage', action=partial(open_web, 'https://flexget.com/'))
        self.add_menu_item(text='Forum', action=partial(open_web, 'https://discuss.flexget.com/'))

    @property
    def menu(self) -> 'Menu':
        # This is lazy loaded since we'd like to delay the menu build until the tray is requested to run
        if not self._menu:
            self._menu = Menu(*self.menu_items)
        return self._menu

    @check_if_tray_is_active
    def run(self):
        """Run the tray icon. Must be run from the main thread and is blocking"""
        try:
            logger.verbose('Starting tray icon')
            self.icon = Icon('Flexget', Image.open(self.path_to_image), menu=self.menu)
            self.running = True
            self.icon.run()
        except Exception as e:
            logger.warning('Could not run tray icon: {}', e)
            self.running = False

    @check_if_tray_is_active
    def stop(self):
        if not self.running:
            return

        logger.verbose('Stopping tray icon')
        self.icon.stop()
        self.running = False


tray_icon = TrayIcon()



from loguru import logger

from flexget import options
from flexget.event import event
from flexget.manager import Session
from flexget.terminal import console

logger = logger.bind(name='perftests')

TESTS = ['imdb_query']


def cli_perf_test(manager, options):
    if options.test_name not in TESTS:
        console('Unknown performance test %s' % options.test_name)
        return
    session = Session()
    try:
        if options.test_name == 'imdb_query':
            imdb_query(session)
    finally:
        session.close()


def imdb_query(session):
    import time

    from progressbar import ETA, Bar, Percentage, ProgressBar
    from sqlalchemy.orm import joinedload_all
    from sqlalchemy.sql.expression import select

    # NOTE: importing other plugins directly is discouraged
    from flexget.components.imdb.db import Movie
    from flexget.plugins.cli.performance import log_query_count

    imdb_urls = []

    logger.info('Getting imdb_urls ...')
    # query so that we avoid loading whole object (maybe cached?)
    for _, url in session.execute(select([Movie.id, Movie.url])):
        imdb_urls.append(url)
    logger.info('Got {} urls from database', len(imdb_urls))
    if not imdb_urls:
        logger.info('so .. aborting')
        return

    # commence testing

    widgets = ['Benchmarking - ', ETA(), ' ', Percentage(), ' ', Bar(left='[', right=']')]
    bar = ProgressBar(widgets=widgets, maxval=len(imdb_urls)).start()

    log_query_count('test')
    start_time = time.time()
    for index, url in enumerate(imdb_urls):
        bar.update(index)

        # movie = session.query(Movie).filter(Movie.url == url).first()
        # movie = session.query(Movie).options(subqueryload(Movie.genres)).filter(Movie.url == url).one()

        movie = (
            session.query(Movie)
            .options(joinedload_all(Movie.genres, Movie.languages, Movie.actors, Movie.directors))
            .filter(Movie.url == url)
            .first()
        )

        # access it's members so they're loaded
        [x.name for x in movie.genres]
        [x.name for x in movie.directors]
        [x.name for x in movie.actors]
        [x.name for x in movie.languages]

    log_query_count('test')
    took = time.time() - start_time
    logger.debug('Took %.2f seconds to query %i movies' % (took, len(imdb_urls)))


@event('options.register')
def register_parser_arguments():
    perf_parser = options.register_command('perf-test', cli_perf_test)
    perf_parser.add_argument('test_name', metavar='<test name>', choices=TESTS)



"""Input plugin for www.betaseries.com"""
from hashlib import md5

from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests
from flexget.utils.cached_input import cached

logger = logger.bind(name='betaseries_list')

API_URL_PREFIX = 'https://api.betaseries.com/'


class BetaSeriesList:
    """
    Emits an entry for each serie followed by one or more BetaSeries account.
    See https://www.betaseries.com/

    Configuration examples:

    # will get all series followed by the account identified by your_user_name
    betaseries_list:
      username: your_user_name
      password: your_password
      api_key: your_api_key

    # will get all series followed by the account identified by some_other_guy
    betaseries_list:
      username: your_user_name
      password: your_password
      api_key: your_api_key
      members:
        - some_other_guy

    # will get all series followed by the accounts identified by guy1 and guy2
    betaseries_list:
      username: your_user_name
      password: your_password
      api_key: your_api_key
      members:
        - guy1
        - guy2


    Api key can be requested at https://www.betaseries.com/api.

    This plugin is meant to work with the import_series plugin as follow:

    import_series:
      from:
        betaseries_list:
          username: xxxxx
          password: xxxxx
          api_key: xxxxx
    """

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'api_key': {'type': 'string'},
            'members': {'type': 'array', 'items': {"title": 'member name', "type": "string"}},
        },
        'required': ['username', 'password', 'api_key'],
        'additionalProperties': False,
    }

    @cached('betaseries_list', persist='2 hours')
    def on_task_input(self, task, config):
        username = config['username']
        password = config['password']
        api_key = config['api_key']
        members = config.get('members', [username])

        titles = set()
        try:
            user_token = create_token(api_key, username, password)
            for member in members:
                titles.update(query_series(api_key, user_token, member))
        except (requests.RequestException, AssertionError) as err:
            logger.opt(exception=True).critical(
                'Failed to get series at BetaSeries.com: {}', err.message
            )

        logger.verbose('series: ' + ', '.join(titles))
        entries = []
        for t in titles:
            e = Entry()
            e['title'] = t
            entries.append(e)
        return entries


def create_token(api_key, login, password):
    """
    login in and request an new API token.
    https://www.betaseries.com/wiki/Documentation#cat-members

    :param string api_key: Api key requested at https://www.betaseries.com/api
    :param string login: Login name
    :param string password: Password
    :return: User token
    """
    r = requests.post(
        API_URL_PREFIX + 'members/auth',
        params={'login': login, 'password': md5(password.encode('utf-8')).hexdigest()},
        headers={
            'Accept': 'application/json',
            'X-BetaSeries-Version': '2.1',
            'X-BetaSeries-Key': api_key,
        },
    )
    assert r.status_code == 200, "Bad HTTP status code: %s" % r.status_code
    j = r.json()
    error_list = j['errors']
    for err in error_list:
        logger.error(str(err))
    if not error_list:
        return j['token']


def query_member_id(api_key, user_token, login_name):
    """
    Get the member id of a member identified by its login name.

    :param string api_key: Api key requested at https://www.betaseries.com/api
    :param string user_token: obtained with a call to create_token()
    :param string login_name: The login name of the member
    :return: Id of the member identified by its login name or `None` if not found
    """
    r = requests.get(
        API_URL_PREFIX + 'members/search',
        params={'login': login_name},
        headers={
            'Accept': 'application/json',
            'X-BetaSeries-Version': '2.1',
            'X-BetaSeries-Key': api_key,
            'X-BetaSeries-Token': user_token,
        },
    )
    assert r.status_code == 200, "Bad HTTP status code: %s" % r.status_code
    j = r.json()
    error_list = j['errors']
    for err in error_list:
        logger.error(str(err))
    found_id = None
    if not error_list:
        for candidate in j['users']:
            if candidate['login'] == login_name:
                found_id = candidate['id']
                break
    return found_id


def query_series(api_key, user_token, member_name=None):
    """
    Get the list of series followed by the authenticated user

    :param string api_key: Api key requested at https://www.betaseries.com/api
    :param string user_token: Obtained with a call to create_token()
    :param string member_name: [optional] A member name to get the list of series from. If None, will query the member
        for whom the user_token was for
    :return: List of serie titles or empty list
    """
    params = {}
    if member_name:
        member_id = query_member_id(api_key, user_token, member_name)
        if member_id:
            params = {'id': member_id}
        else:
            logger.error('member {!r} not found', member_name)
            return []
    r = requests.get(
        API_URL_PREFIX + 'shows/member',
        params=params,
        headers={
            'Accept': 'application/json',
            'X-BetaSeries-Version': '2.1',
            'X-BetaSeries-Key': api_key,
            'X-BetaSeries-Token': user_token,
        },
    )
    assert r.status_code == 200, "Bad HTTP status code: %s" % r.status_code
    j = r.json()
    error_list = j['errors']
    for err in error_list:
        logger.error(str(err))
    if not error_list:
        return [x['title'] for x in j['shows'] if x['user']['archived'] is False]
    else:
        return []


@event('plugin.register')
def register_plugin():
    plugin.register(BetaSeriesList, 'betaseries_list', api_ver=2, interfaces=['task'])



from fnmatch import fnmatch

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event

logger = logger.bind(name='content_filter')


class FilterContentFilter:
    """
    Rejects entries based on the filenames in the content. Torrent files only right now.

    Example::

      content_filter:
        require:
          - '*.avi'
          - '*.mkv'
    """

    schema = {
        'type': 'object',
        'properties': {
            # These two properties allow a string or list of strings
            'require': one_or_more({'type': 'string'}),
            'require_all': one_or_more({'type': 'string'}),
            'reject': one_or_more({'type': 'string'}),
            'require_mainfile': {'type': 'boolean', 'default': False},
            'strict': {'type': 'boolean', 'default': False},
        },
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        for key in ['require', 'require_all', 'reject']:
            if key in config:
                if isinstance(config[key], str):
                    config[key] = [config[key]]
        return config

    def process_entry(self, task, entry, config):
        """
        Process an entry and reject it if it doesn't pass filter.

        :param task: Task entry belongs to.
        :param entry: Entry to process
        :return: True, if entry was rejected.
        """
        if 'content_files' in entry:
            files = entry['content_files']
            logger.debug('{} files: {}', entry['title'], files)

            def matching_mask(files, masks):
                """Returns matching mask if any files match any of the masks, false otherwise"""
                for file in files:
                    for mask in masks:
                        if fnmatch(file, mask):
                            return mask
                return False

            # Avoid confusion by printing a reject message to info log, as
            # download plugin has already printed a downloading message.
            if config.get('require'):
                if not matching_mask(files, config['require']):
                    logger.info(
                        'Entry {} does not have any of the required filetypes, rejecting',
                        entry['title'],
                    )
                    entry.reject('does not have any of the required filetypes', remember=True)
                    return True
            if config.get('require_all'):
                # Make sure each mask matches at least one of the contained files
                if not all(
                    any(fnmatch(file, mask) for file in files) for mask in config['require_all']
                ):
                    logger.info(
                        'Entry {} does not have all of the required filetypes, rejecting',
                        entry['title'],
                    )
                    entry.reject('does not have all of the required filetypes', remember=True)
                    return True
            if config.get('reject'):
                mask = matching_mask(files, config['reject'])
                if mask:
                    logger.info('Entry {} has banned file {}, rejecting', entry['title'], mask)
                    entry.reject('has banned file %s' % mask, remember=True)
                    return True
            if config.get('require_mainfile') and len(files) > 1:
                best = None
                for f in entry['torrent'].get_filelist():
                    if not best or f['size'] > best:
                        best = f['size']
                if (100 * float(best) / float(entry['torrent'].size)) < 90:
                    logger.info('Entry {} does not have a main file, rejecting', entry['title'])
                    entry.reject('does not have a main file', remember=True)
                    return True

    @plugin.priority(150)
    def on_task_modify(self, task, config):
        if task.options.test or task.options.learn:
            logger.info(
                'Plugin is partially disabled with --test and --learn '
                'because content filename information may not be available'
            )
            # return

        config = self.prepare_config(config)
        for entry in task.accepted:
            if self.process_entry(task, entry, config):
                task.rerun(plugin='content_filter')
            elif 'content_files' not in entry and config.get('strict'):
                entry.reject('no content files parsed for entry', remember=True)
                task.rerun(plugin='content_filter')


@event('plugin.register')
def register_plugin():
    plugin.register(FilterContentFilter, 'content_filter', api_ver=2)



from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode, and_, func
from sqlalchemy.orm import relationship

from flexget.db_schema import versioned_base, with_session
from flexget.entry import Entry

logger = logger.bind(name='regexp_list')
Base = versioned_base('regexp_list', 1)


class RegexpListList(Base):
    __tablename__ = 'regexp_list_lists'
    id = Column(Integer, primary_key=True)
    name = Column(Unicode, unique=True)
    added = Column(DateTime, default=datetime.now)
    regexps = relationship(
        'RegexListRegexp', backref='list', cascade='all, delete, delete-orphan', lazy='dynamic'
    )

    def __repr__(self):
        return '<RegexpListList name=%s,id=%d>' % (self.name, self.id)

    def to_dict(self):
        return {'id': self.id, 'name': self.name, 'added_on': self.added}


class RegexListRegexp(Base):
    __tablename__ = 'regexp_list_regexps'
    id = Column(Integer, primary_key=True)
    added = Column(DateTime, default=datetime.now)
    regexp = Column(Unicode)
    list_id = Column(Integer, ForeignKey(RegexpListList.id), nullable=False)

    def __repr__(self):
        return '<RegexListRegexp regexp=%s,list_name=%s>' % (self.regexp, self.list.name)

    def to_entry(self):
        entry = Entry()
        entry['title'] = entry['regexp'] = self.regexp
        entry['url'] = 'mock://localhost/regexp_list/%d' % self.id
        return entry

    def to_dict(self):
        return {'id': self.id, 'added_on': self.added, 'regexp': self.regexp}


@with_session
def get_regexp_lists(name=None, session=None):
    logger.debug('retrieving regexp lists')
    query = session.query(RegexpListList)
    if name:
        logger.debug('filtering by name {}', name)
        query = query.filter(RegexpListList.name.contains(name))
    return query.all()


@with_session
def get_list_by_exact_name(name, session=None):
    logger.debug('returning list with name {}', name)
    return (
        session.query(RegexpListList)
        .filter(func.lower(RegexpListList.name) == name.lower())
        .one_or_none()
    )


@with_session
def get_regexps_by_list_id(
    list_id, count=False, start=None, stop=None, order_by='added', descending=False, session=None
):
    query = session.query(RegexListRegexp).filter(RegexListRegexp.list_id == list_id)
    if count:
        return query.count()
    query = query.slice(start, stop).from_self()
    if descending:
        query = query.order_by(getattr(RegexListRegexp, order_by).desc())
    else:
        query = query.order_by(getattr(RegexListRegexp, order_by))
    return query.all()


@with_session
def get_list_by_id(list_id, session=None):
    logger.debug('fetching list with id {}', list_id)
    return session.query(RegexpListList).filter(RegexpListList.id == list_id).one_or_none()


@with_session
def get_regexp(list_id, regexp, session=None):
    regexp_list = get_list_by_id(list_id=list_id, session=session)
    if regexp_list:
        logger.debug('searching for regexp {} in list {}', regexp, list_id)
        return (
            session.query(RegexListRegexp)
            .filter(
                and_(
                    func.lower(RegexListRegexp.regexp) == regexp.lower(),
                    RegexListRegexp.list_id == list_id,
                )
            )
            .first()
        )


@with_session
def create_list(list_name, session=None):
    """
    Only creates the list if it doesn't exist.

    :param str list_name: Name of the list
    :param Session session:
    :return: regex list with name list_name
    """
    regexp_list = get_list_by_exact_name(list_name, session=session)
    if not regexp_list:
        regexp_list = RegexpListList(name=list_name)
        session.merge(regexp_list)
        session.commit()
    return regexp_list


@with_session
def add_to_list_by_name(list_name, regexp, session=None):
    regexp_list = create_list(list_name, session=session)
    existing_regexp = get_regexp(regexp_list.id, regexp, session=session)
    if not existing_regexp:
        new_regexp = RegexListRegexp(regexp=regexp, list_id=regexp_list.id)
        session.merge(new_regexp)
        session.commit()



from collections.abc import MutableSet

from loguru import logger
from sqlalchemy import func
from sqlalchemy.sql.elements import and_

from flexget import plugin
from flexget.db_schema import with_session
from flexget.event import event
from flexget.manager import Session
from flexget.utils.tools import split_title_year

from . import db

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.parsing.parsers import parser_common as plugin_parser_common
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='parser_common')

logger = logger.bind(name='movie_list')


class MovieListBase:
    """
    Class that contains helper methods for movie list as well as plugins that use it,
    such as API and CLI.
    """

    @property
    def supported_ids(self):
        # Return a list of supported series identifier as registered via their plugins
        return [
            p.instance.movie_identifier for p in plugin.get_plugins(interface='movie_metainfo')
        ]


class MovieList(MutableSet):
    def _db_list(self, session):
        return (
            session.query(db.MovieListList).filter(db.MovieListList.name == self.list_name).first()
        )

    def _from_iterable(self, it):
        # TODO: is this the right answer? the returned object won't have our custom __contains__ logic
        return set(it)

    @with_session
    def __init__(self, config, session=None):
        if not isinstance(config, dict):
            config = {'list_name': config}
        config.setdefault('strip_year', False)
        self.list_name = config.get('list_name')
        self.strip_year = config.get('strip_year')

        db_list = self._db_list(session)
        if not db_list:
            session.add(db.MovieListList(name=self.list_name))

    def __iter__(self):
        with Session() as session:
            return iter(
                [movie.to_entry(self.strip_year) for movie in self._db_list(session).movies]
            )

    def __len__(self):
        with Session() as session:
            return len(self._db_list(session).movies)

    def add(self, entry):
        with Session() as session:
            # Check if this is already in the list, refresh info if so
            db_list = self._db_list(session=session)
            db_movie = self._find_entry(entry, session=session)
            # Just delete and re-create to refresh
            if db_movie:
                session.delete(db_movie)
            db_movie = db.MovieListMovie()
            if 'movie_name' in entry:
                db_movie.title, db_movie.year = entry['movie_name'], entry.get('movie_year')
            else:
                db_movie.title, db_movie.year = split_title_year(entry['title'])
            for id_name in MovieListBase().supported_ids:
                if id_name in entry:
                    db_movie.ids.append(db.MovieListID(id_name=id_name, id_value=entry[id_name]))
            logger.debug('adding entry {}', entry)
            db_list.movies.append(db_movie)
            session.commit()
            return db_movie.to_entry()

    def discard(self, entry):
        with Session() as session:
            db_movie = self._find_entry(entry, session=session)
            if db_movie:
                logger.debug('deleting movie {}', db_movie)
                session.delete(db_movie)

    def __contains__(self, entry):
        return self._find_entry(entry) is not None

    @with_session
    def _find_entry(self, entry, session=None):
        """Finds `MovieListMovie` corresponding to this entry, if it exists."""
        # Match by supported IDs
        for id_name in MovieListBase().supported_ids:
            if entry.get(id_name):
                logger.debug('trying to match movie based off id {}: {}', id_name, entry[id_name])
                res = (
                    self._db_list(session)
                    .movies.join(db.MovieListMovie.ids)
                    .filter(
                        and_(
                            db.MovieListID.id_name == id_name,
                            db.MovieListID.id_value == entry[id_name],
                        )
                    )
                    .first()
                )
                if res:
                    logger.debug('found movie {}', res)
                    return res
        # Fall back to title/year match
        if not entry.get('movie_name'):
            self._parse_title(entry)
        if entry.get('movie_name'):
            name = entry['movie_name']
            year = entry.get('movie_year') if entry.get('movie_year') else None
        else:
            logger.warning('Could not get a movie name, skipping')
            return
        logger.debug('trying to match movie based of name: {} and year: {}', name, year)
        res = (
            self._db_list(session)
            .movies.filter(func.lower(db.MovieListMovie.title) == name.lower())
            .filter(db.MovieListMovie.year == year)
            .first()
        )
        if res:
            logger.debug('found movie {}', res)
        return res

    @staticmethod
    def _parse_title(entry):
        parser = plugin.get('parsing', 'movie_list').parse_movie(data=entry['title'])
        if parser and parser.valid:
            parser.name = plugin_parser_common.normalize_name(
                plugin_parser_common.remove_dirt(parser.name)
            )
            entry.update(parser.fields)

    @property
    def immutable(self):
        return False

    @property
    def online(self):
        """
        Set the online status of the plugin, online plugin should be treated
        differently in certain situations, like test mode
        """
        return False

    @with_session
    def get(self, entry, session):
        match = self._find_entry(entry=entry, session=session)
        return match.to_entry() if match else None


class PluginMovieList:
    """Remove all accepted elements from your trakt.tv watchlist/library/seen or custom list."""

    schema = {
        'oneOf': [
            {'type': 'string'},
            {
                'type': 'object',
                'properties': {'list_name': {'type': 'string'}, 'strip_year': {'type': 'boolean'}},
                'required': ['list_name'],
                'additionalProperties': False,
            },
        ]
    }

    @staticmethod
    def get_list(config):
        return MovieList(config)

    def on_task_input(self, task, config):
        return list(MovieList(config))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginMovieList, 'movie_list', api_ver=2, interfaces=['task', 'list'])



import re
from collections.abc import MutableSet

from loguru import logger

from flexget import plugin
from flexget.db_schema import with_session
from flexget.event import event
from flexget.manager import Session

from . import db

logger = logger.bind(name='regexp_list')


class RegexpList(MutableSet):
    schema = {'type': 'string'}

    def _db_list(self, session):
        return (
            session.query(db.RegexpListList).filter(db.RegexpListList.name == self.config).first()
        )

    def _from_iterable(self, it):
        # TODO: is this the right answer? the returned object won't have our custom __contains__ logic
        return set(it)

    @with_session
    def __init__(self, config, session=None):
        self.config = config
        db_list = self._db_list(session)
        if not db_list:
            session.add(db.RegexpListList(name=self.config))

    def __iter__(self):
        with Session() as session:
            return iter([regexp.to_entry() for regexp in self._db_list(session).regexps])

    def __len__(self):
        with Session() as session:
            return self._db_list(session).regexps.count()

    def add(self, entry):
        with Session() as session:
            # Check if this is already in the list, refresh info if so
            db_list = self._db_list(session=session)
            db_regexp = self._find_entry(entry, session=session)
            # Just delete and re-create to refresh
            if db_regexp:
                session.delete(db_regexp)
            db_regexp = db.RegexListRegexp()
            db_regexp.regexp = entry.get('regexp', entry['title'])
            db_list.regexps.append(db_regexp)
            session.commit()
            return db_regexp.to_entry()

    def discard(self, entry):
        with Session() as session:
            for match_regexp in [False, True]:
                db_regexp = self._find_entry(entry, match_regexp=match_regexp, session=session)
                if db_regexp:
                    logger.debug('deleting file {}', db_regexp)
                    session.delete(db_regexp)

    def __contains__(self, entry):
        return self._find_entry(entry, match_regexp=True) is not None

    @with_session
    def _find_entry(self, entry, match_regexp=False, session=None):
        """Finds `SubtitleListFile` corresponding to this entry, if it exists."""
        res = None
        if match_regexp:
            for regexp in self._db_list(session).regexps:
                if re.search(regexp.regexp, entry['title'], re.IGNORECASE):
                    res = regexp
        else:
            res = (
                self._db_list(session)
                .regexps.filter(db.RegexListRegexp.regexp == entry.get('regexp', entry['title']))
                .first()
            )
        return res

    @property
    def immutable(self):
        return False

    @property
    def online(self):
        """Set the online status of the plugin, online plugin should be treated differently in certain situations,
        like test mode"""
        return False

    @with_session
    def get(self, entry, session):
        match = self._find_entry(entry=entry, match_regexp=True, session=session)
        return match.to_entry() if match else None


class PluginRegexpList:
    """Subtitle list"""

    schema = RegexpList.schema

    @staticmethod
    def get_list(config):
        return RegexpList(config)

    def on_task_input(self, task, config):
        regexp_list = RegexpList(config)

        return list(regexp_list)


@event('plugin.register')
def register_plugin():
    plugin.register(PluginRegexpList, 'regexp_list', api_ver=2, interfaces=['task', 'list'])



from urllib.parse import urlencode

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='sabnzbd')


class OutputSabnzbd:
    """
    Example::

      sabnzbd:
        apikey: 123456
        url: http://localhost/sabnzbd/api?
        category: movies

    All parameters::

      sabnzbd:
        apikey: ...
        url: ...
        category: ...
        script: ...
        pp: ...
        priority: ...
    """

    schema = {
        'type': 'object',
        'properties': {
            'key': {'type': 'string'},
            'url': {'type': 'string', 'format': 'url'},
            'category': {'type': 'string'},
            'script': {'type': 'string'},
            'pp': {'type': 'string'},
            'priority': {'type': 'integer'},
            'password': {'type': 'string'},
            'username': {'type': 'string'},
        },
        'required': ['key', 'url'],
        'additionalProperties': False,
    }

    def get_params(self, config):
        params = {}
        if 'key' in config:
            params['apikey'] = config['key']
        if 'category' in config:
            params['cat'] = '%s' % config['category']
        if 'script' in config:
            params['script'] = config['script']
        if 'pp' in config:
            params['pp'] = config['pp']
        if 'priority' in config:
            params['priority'] = config['priority']
        if 'username' in config:
            params['ma_username'] = config['username']
        if 'password' in config:
            params['ma_password'] = config['password']
        return params

    def on_task_output(self, task, config):
        for entry in task.accepted:
            if task.options.test:
                logger.info('Would add into sabnzbd: {}', entry['title'])
                continue

            params = self.get_params(config)
            # allow overriding the category
            if 'category' in entry:
                # Dirty hack over the next few lines to strip out non-ascii
                # chars. We're going to urlencode this, which causes
                # serious issues in python2.x if it's not ascii input.
                params['cat'] = ''.join([x for x in entry['category'] if ord(x) < 128])
            params['name'] = ''.join([x for x in entry['url'] if ord(x) < 128])
            # add cleaner nzb name (undocumented api feature)
            params['nzbname'] = ''.join([x for x in entry['title'] if ord(x) < 128])

            # check whether file is local or remote
            if entry['url'].startswith('file://'):
                params['mode'] = 'addlocalfile'
                params['name'] = entry['location']
            else:
                params['mode'] = 'addurl'

            request_url = config['url'] + urlencode(params)
            logger.debug('request_url: {}', request_url)
            try:
                response = task.requests.get(request_url)
            except RequestException as e:
                logger.critical('Failed to use sabnzbd. Requested {}', request_url)
                logger.critical('Result was: {}', e.args[0])
                entry.fail('sabnzbd unreachable')
                if task.options.debug:
                    logger.exception(e)
                continue

            if 'error' in response.text.lower():
                entry.fail(response.text.replace('\n', ''))
            else:
                logger.info('Added `{}` to SABnzbd', entry['title'])


@event('plugin.register')
def register_plugin():
    plugin.register(OutputSabnzbd, 'sabnzbd', api_ver=2)



import datetime
import re
from string import capwords
from typing import List, Optional, Tuple, Union

from loguru import logger

from flexget.utils.qualities import Quality

logger = logger.bind(name='parser')

SERIES_ID_TYPES = ['ep', 'date', 'sequence', 'id']


def clean_value(name: str) -> str:
    for char in '[]()_,.':
        name = name.replace(char, ' ')

    # if there are no spaces
    if name.find(' ') == -1:
        name = name.replace('-', ' ')

    # MovieParser.strip_spaces
    name = ' '.join(name.split())
    return name


def old_assume_quality(guessed_quality: Quality, assumed_quality: Quality) -> Quality:
    if assumed_quality:
        if not guessed_quality:
            return assumed_quality
        if assumed_quality.resolution:
            guessed_quality.resolution = assumed_quality.resolution
        if assumed_quality.source:
            guessed_quality.source = assumed_quality.source
        if assumed_quality.codec:
            guessed_quality.codec = assumed_quality.codec
        if assumed_quality.audio:
            guessed_quality.audio = assumed_quality.audio
    return guessed_quality


def remove_dirt(name: str) -> str:
    if name:
        name = re.sub(r'[_.,\[\]\(\): ]+', ' ', name).strip().lower()
    return name


def normalize_name(name: str) -> str:
    name = capwords(name)
    return name


class MovieParseResult:
    def __init__(
        self,
        data: str = None,
        name: str = None,
        year: Optional[int] = None,
        quality: Quality = None,
        proper_count: int = 0,
        release_group: Optional[str] = None,
        valid: bool = True,
    ) -> None:
        self.name: str = name
        self.data: str = data
        self.year: Optional[int] = year
        self.quality: Quality = quality if quality is not None else Quality()
        self.proper_count: int = proper_count
        self.release_group: Optional[str] = release_group
        self.valid: bool = valid

    @property
    def identifier(self) -> str:
        if self.name and self.year:
            return ('%s %s' % (self.name, self.year)).strip().lower()
        elif self.name:
            return self.name.lower()

    @property
    def proper(self) -> bool:
        return self.proper_count > 0

    @property
    def fields(self) -> dict:
        """
        Return a dict of all parser fields
        """
        return {
            'id': self.identifier,
            'movie_parser': self,
            'movie_name': self.name,
            'movie_year': self.year,
            'proper': self.proper,
            'proper_count': self.proper_count,
            'release_group': self.release_group,
        }

    def __str__(self) -> str:
        valid = 'OK' if self.valid else 'INVALID'
        return '<MovieParseResult(data=%s,name=%s,year=%s,id=%s,quality=%s,proper=%s,release_group=%s,status=%s)>' % (
            self.data,
            self.name,
            self.year,
            self.identifier,
            self.quality,
            self.proper_count,
            self.release_group,
            valid,
        )


class SeriesParseResult:
    def __init__(
        self,
        data: str = None,
        name: str = None,
        identified_by: str = None,
        id_type: str = None,
        id: Union[Tuple[int, int], str, int, datetime.date] = None,
        episodes: int = 1,
        season_pack: bool = False,
        strict_name: bool = False,
        quality: Quality = None,
        proper_count: int = 0,
        special: bool = False,
        group: Optional[str] = None,
        valid: bool = True,
    ) -> None:
        self.name: str = name
        self.data: str = data
        self.episodes: int = episodes
        self.season_pack: bool = season_pack
        self.identified_by: str = identified_by
        self.id: Union[Tuple[int, int], str, int, datetime.date] = id
        self.id_type: str = id_type
        self.quality: Quality = quality if quality is not None else Quality()
        self.proper_count: int = proper_count
        self.special: bool = special
        self.group: Optional[str] = group
        self.valid: bool = valid
        self.strict_name: bool = strict_name

    @property
    def proper(self) -> bool:
        return self.proper_count > 0

    @property
    def season(self) -> Optional[int]:
        if self.id_type == 'ep':
            return self.id[0]
        if self.id_type == 'date':
            return self.id.year
        if self.id_type == 'sequence':
            return 0
        return None

    @property
    def episode(self) -> Optional[int]:
        if self.id_type == 'ep':
            return self.id[1]
        if self.id_type == 'sequence':
            return self.id
        return None

    @property
    def identifiers(self) -> List[str]:
        """Return all identifiers this parser represents. (for packs)"""
        # Currently 'ep' is the only id type that supports packs
        if not self.valid:
            raise Exception('Series flagged invalid')
        if self.id_type == 'ep':
            if self.season_pack:
                return ['S%02d' % self.season]
            return ['S%02dE%02d' % (self.season, self.episode + x) for x in range(self.episodes)]
        elif self.id_type == 'date':
            return [self.id.strftime('%Y-%m-%d')]
        if self.id is None:
            raise Exception('Series is missing identifier')
        else:
            return [self.id]

    @property
    def identifier(self) -> str:
        """Return String identifier for parsed episode, eg. S01E02
        (will be the first identifier if this is a pack)
        """
        return self.identifiers[0]

    @property
    def pack_identifier(self) -> str:
        """Return a combined identifier for the whole pack if this has more than one episode."""
        # Currently only supports ep mode
        if self.id_type == 'ep':
            if self.episodes > 1:
                return 'S%02dE%02d-E%02d' % (
                    self.season,
                    self.episode,
                    self.episode + self.episodes - 1,
                )
            else:
                return self.identifier
        else:
            return self.identifier

    def __str__(self) -> str:
        valid = 'OK' if self.valid else 'INVALID'
        return '<SeriesParseResult(data=%s,name=%s,id=%s,season=%s,season_pack=%s,episode=%s,quality=%s,proper=%s,' 'special=%s,status=%s)>' % (
            self.data,
            self.name,
            str(self.id),
            self.season,
            self.season_pack,
            self.episode,
            self.quality,
            self.proper_count,
            self.special,
            valid,
        )



from collections.abc import MutableSet
from urllib.parse import urlparse

import requests
from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='couchpotato_list')


class CouchPotatoBase:
    @staticmethod
    def movie_list_request(base_url, port, api_key):
        parsedurl = urlparse(base_url)
        logger.debug('Received movie list request')
        return '%s://%s:%s%s/api/%s/movie.list?status=active' % (
            parsedurl.scheme,
            parsedurl.netloc,
            port,
            parsedurl.path,
            api_key,
        )

    @staticmethod
    def profile_list_request(base_url, port, api_key):
        parsedurl = urlparse(base_url)
        logger.debug('Received profile list request')
        return '%s://%s:%s%s/api/%s/profile.list' % (
            parsedurl.scheme,
            parsedurl.netloc,
            port,
            parsedurl.path,
            api_key,
        )

    @staticmethod
    def movie_add_request(base_url, port, api_key):
        parsedurl = urlparse(base_url)
        logger.debug('Received movie add request')
        return '%s://%s:%s%s/api/%s/movie.add' % (
            parsedurl.scheme,
            parsedurl.netloc,
            port,
            parsedurl.path,
            api_key,
        )

    @staticmethod
    def movie_delete_request(base_url, port, api_key):
        parsedurl = urlparse(base_url)
        logger.debug('Received movie delete request')
        return '%s://%s:%s%s/api/%s/movie.delete?delete_from=wanted' % (
            parsedurl.scheme,
            parsedurl.netloc,
            port,
            parsedurl.path,
            api_key,
        )

    @staticmethod
    def build_url(base_url, request_type, port, api_key):
        if request_type == 'active':
            return CouchPotatoBase.movie_list_request(base_url, port, api_key)
        elif request_type == 'profiles':
            return CouchPotatoBase.profile_list_request(base_url, port, api_key)
        elif request_type == 'add':
            return CouchPotatoBase.movie_add_request(base_url, port, api_key)
        elif request_type == 'delete':
            return CouchPotatoBase.movie_delete_request(base_url, port, api_key)
        else:
            raise plugin.PluginError('Received unknown API request, aborting.')

    @staticmethod
    def get_json(url):
        try:
            return requests.get(url).json()
        except RequestException as e:
            raise plugin.PluginError(
                'Unable to connect to Couchpotato at %s. Error: %s' % (url, e)
            )

    @staticmethod
    def quality_requirement_builder(quality_profile):
        """
        Converts CP's quality profile to a format that can be converted to FlexGet QualityRequirement
        """
        # TODO: Not all values have exact matches in flexget, need to update flexget qualities
        sources = {
            'BR-Disk': 'remux',  # Not a perfect match, but as close as currently possible
            'brrip': 'bluray',
            'dvdr': 'dvdrip',  # Not a perfect match, but as close as currently possible
            'dvdrip': 'dvdrip',
            'scr': 'dvdscr',
            'r5': 'r5',
            'tc': 'tc',
            'ts': 'ts',
            'cam': 'cam',
        }

        resolutions = {'1080p': '1080p', '720p': '720p'}

        # Separate strings are needed for each QualityComponent
        # TODO list is converted to set because if a quality has 3d type in CP, it gets duplicated during the conversion
        # TODO when (and if) 3d is supported in flexget this will be needed to removed
        res_string = '|'.join(
            set(
                [
                    resolutions[quality]
                    for quality in quality_profile['qualities']
                    if quality in resolutions
                ]
            )
        )
        source_string = '|'.join(
            set(
                [
                    sources[quality]
                    for quality in quality_profile['qualities']
                    if quality in sources
                ]
            )
        )

        quality_requirement = (res_string + ' ' + source_string).rstrip()
        logger.debug('quality requirement is {}', quality_requirement)
        return quality_requirement

    @staticmethod
    def list_entries(config, test_mode=None):
        logger.verbose('Connecting to CouchPotato to retrieve movie list.')
        active_movies_url = CouchPotatoBase.build_url(
            config.get('base_url'), 'active', config.get('port'), config.get('api_key')
        )
        active_movies_json = CouchPotatoBase.get_json(active_movies_url)
        # Gets profile and quality lists if include_data is TRUE
        if config.get('include_data'):
            logger.verbose('Connecting to CouchPotato to retrieve profile data.')
            profile_url = CouchPotatoBase.build_url(
                config.get('base_url'), 'profiles', config.get('port'), config.get('api_key')
            )
            profile_json = CouchPotatoBase.get_json(profile_url)

        entries = []
        for movie in active_movies_json['movies']:
            # Related to #1444, corrupt data from CP
            if not all([movie.get('status'), movie.get('title'), movie.get('info')]):
                logger.warning('corrupt movie data received, skipping')
                continue
            quality_req = ''
            logger.debug('movie data: {}', movie)
            if movie['status'] == 'active':
                if config.get('include_data') and profile_json:
                    for profile in profile_json['list']:
                        if (
                            profile['_id'] == movie['profile_id']
                        ):  # Matches movie profile with profile JSON
                            quality_req = CouchPotatoBase.quality_requirement_builder(profile)
                entry = Entry(
                    title=movie["title"],
                    url='',
                    imdb_id=movie['info'].get('imdb'),
                    tmdb_id=movie['info'].get('tmdb_id'),
                    quality_req=quality_req,
                    couchpotato_id=movie.get('_id'),
                )
                if entry.isvalid():
                    logger.debug('returning entry {}', entry)
                    entries.append(entry)
                else:
                    logger.error('Invalid entry created? {}', entry)
                    continue
                # Test mode logging
                if entry and test_mode:
                    logger.info("Test mode. Entry includes:")
                    for key, value in entry.items():
                        logger.info('     {}: {}', key.capitalize(), value)

        return entries

    @staticmethod
    def add_movie(config, entry, test_mode=None):
        if not entry.get('imdb_id'):
            logger.error('Cannot add movie to couchpotato without an imdb ID: {}', entry)
            return
        logger.verbose('Connection to CouchPotato to add a movie to list.')
        add_movie_url = CouchPotatoBase.build_url(
            config.get('base_url'), 'add', config.get('port'), config.get('api_key')
        )
        title = entry.get('movie_name')
        imdb_id = entry.get('imdb_id')
        add_movie_url += '?title=%s&identifier=%s' % (title, imdb_id)
        add_movie_json = CouchPotatoBase.get_json(add_movie_url)
        return add_movie_json['movie']

    @staticmethod
    def remove_movie(config, movie_id, test_mode=None):
        logger.verbose('Deleting movie from Couchpotato')
        delete_movie_url = CouchPotatoBase.build_url(
            config.get('base_url'), 'delete', config.get('port'), config.get('api_key')
        )
        delete_movie_url += '&id=%s' % movie_id
        CouchPotatoBase.get_json(delete_movie_url)


class CouchPotatoSet(MutableSet):
    supported_ids = ['couchpotato_id', 'imdb_id', 'tmdb_id']
    schema = {
        'type': 'object',
        'properties': {
            'base_url': {'type': 'string'},
            'port': {'type': 'number', 'default': 80},
            'api_key': {'type': 'string'},
            'include_data': {'type': 'boolean', 'default': False},
        },
        'required': ['api_key', 'base_url'],
        'additionalProperties': False,
    }

    @property
    def movies(self):
        if not self._movies:
            self._movies = CouchPotatoBase.list_entries(self.config)
        return self._movies

    def _find_entry(self, entry):
        for cp_entry in self.movies:
            for sup_id in self.supported_ids:
                if (
                    entry.get(sup_id) is not None
                    and entry[sup_id] == cp_entry[sup_id]
                    or entry.get('title').lower() == cp_entry.get('title').lower()
                ):
                    return cp_entry

    def __init__(self, config):
        self.config = config
        self._movies = None

    def __iter__(self):
        return (entry for entry in self.movies)

    def __len__(self):
        return len(self.movies)

    def __contains__(self, entry):
        return self._find_entry(entry) is not None

    def add(self, entry):
        if not self._find_entry(entry):
            self._movies = None
            movie = CouchPotatoBase.add_movie(self.config, entry)
            logger.verbose(
                'Successfully added movie {} to CouchPotato', movie['info']['original_title']
            )
        else:
            logger.debug('entry {} already exists in couchpotato list', entry)

    def discard(self, entry):
        for movie in self.movies:
            title = entry.get('movie_name') or entry.get('title')
            if movie.get('title').lower() == title.lower():
                movie_id = movie.get('couchpotato_id')
                logger.verbose('Trying to remove movie {} from CouchPotato', title)
                CouchPotatoBase.remove_movie(self.config, movie_id)
                self._movies = None

    @property
    def immutable(self):
        return False

    @property
    def online(self):
        """Set the online status of the plugin, online plugin should be treated differently in certain situations,
        like test mode"""
        return True

    def get(self, entry):
        return self._find_entry(entry)


class CouchPotatoList:
    schema = CouchPotatoSet.schema

    @staticmethod
    def get_list(config):
        return CouchPotatoSet(config)

    def on_task_input(self, task, config):
        return list(CouchPotatoSet(config))


@event('plugin.register')
def register_plugin():
    plugin.register(CouchPotatoList, 'couchpotato_list', api_ver=2, interfaces=['task', 'list'])



# -*- coding: utf-8 -*-


from dateutil.parser import parse as dateutil_parse
from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.tools import TimedDict

from . import db

logger = logger.bind(name=__name__)

USER_CACHE_DURATION = '15 minutes'  # cache duration for sync data eg. history, collection


class TraktUserCache(TimedDict):
    def __init__(self, cache_time):
        super().__init__(cache_time=cache_time)
        self.updaters = {
            'collection': self._update_collection_cache,
            'watched': self._update_watched_cache,
            'ratings': self._update_ratings_cache,
        }

    def _get_user_cache(self, username=None, account=None):
        identifier = '{}|{}'.format(account, username or 'me')
        self.setdefault(identifier, {}).setdefault('watched', {}).setdefault('shows', {})
        self.setdefault(identifier, {}).setdefault('watched', {}).setdefault('movies', {})
        self.setdefault(identifier, {}).setdefault('collection', {}).setdefault('shows', {})
        self.setdefault(identifier, {}).setdefault('collection', {}).setdefault('movies', {})
        self.setdefault(identifier, {}).setdefault('ratings', {}).setdefault('shows', {})
        self.setdefault(identifier, {}).setdefault('ratings', {}).setdefault('seasons', {})
        self.setdefault(identifier, {}).setdefault('ratings', {}).setdefault('episodes', {})
        self.setdefault(identifier, {}).setdefault('ratings', {}).setdefault('movies', {})

        return self[identifier]

    def _update_collection_cache(self, cache, media_type, username=None, account=None):
        collection = db.get_user_data(
            data_type='collection',
            media_type=media_type,
            session=db.get_session(account),
            username=username,
        )

        for media in collection:
            media_id = media['ids']['trakt']
            cache[media_id] = media
            collected_at = media.get('collected_at') or media.get('last_collected_at')
            cache[media_id]['collected_at'] = dateutil_parse(collected_at, ignoretz=True)

    def _update_watched_cache(self, cache, media_type, username=None, account=None):
        watched = db.get_user_data(
            data_type='watched',
            media_type=media_type,
            session=db.get_session(account),
            username=username,
        )
        for media in watched:
            media_id = media['ids']['trakt']
            cache[media_id] = media
            cache[media_id]['watched_at'] = dateutil_parse(media['last_watched_at'], ignoretz=True)
            cache[media_id]['plays'] = media['plays']

    def _update_ratings_cache(self, cache, media_type, username=None, account=None):
        ratings = db.get_user_data(
            data_type='ratings',
            media_type=media_type,
            session=db.get_session(account=account),
            username=username,
        )
        for media in ratings:
            # get the proper cache from the type returned by trakt
            media_id = media['ids']['trakt']
            cache[media_id] = media
            cache[media_id]['rated_at'] = dateutil_parse(media['rated_at'], ignoretz=True)
            cache[media_id]['rating'] = media['rating']

    def _get_data(self, data_type, media_type, username=None, account=None):
        cache = self._get_user_cache(username=username, account=account)[data_type][media_type]

        if not cache:
            logger.debug('No {} found in cache. Refreshing.', data_type)
            self.updaters[data_type](cache, media_type, username=username, account=account)
            cache = self._get_user_cache(username=username, account=account)[data_type][media_type]

        if not cache:
            logger.warning('No {} data returned from trakt.', data_type)

        return cache

    def get_shows_collection(self, username=None, account=None):
        return self._get_data('collection', 'shows', username=username, account=account)

    def get_movie_collection(self, username=None, account=None):
        return self._get_data('collection', 'movies', username=username, account=account)

    def get_watched_shows(self, username=None, account=None):
        return self._get_data('watched', 'shows', username=username, account=account)

    def get_watched_movies(self, username=None, account=None):
        return self._get_data('watched', 'movies', username=username, account=account)

    def get_show_user_ratings(self, username=None, account=None):
        return self._get_data('ratings', 'shows', username=username, account=account)

    def get_season_user_ratings(self, username=None, account=None):
        return self._get_data('ratings', 'seasons', username=username, account=account)

    def get_episode_user_ratings(self, username=None, account=None):
        return self._get_data('ratings', 'episodes', username=username, account=account)

    def get_movie_user_ratings(self, username=None, account=None):
        return self._get_data('ratings', 'movies', username=username, account=account)


# Global user cache
# TODO better idea?
user_cache = TraktUserCache(cache_time=USER_CACHE_DURATION)


class ApiTrakt:
    def __init__(self, username=None, account=None):
        self.account = account
        self.username = db.get_username(username, account)

    @property
    def lookup_map(self):
        return {
            'watched': {
                'show': self.is_show_watched,
                'season': self.is_season_watched,
                'episode': self.is_episode_watched,
                'movie': self.is_movie_watched,
            },
            'collected': {
                'show': self.is_show_in_collection,
                'season': self.is_season_in_collection,
                'episode': self.is_episode_in_collection,
                'movie': self.is_movie_in_collection,
            },
            'ratings': {
                'show': self.show_user_ratings,
                'season': self.season_user_ratings,
                'episode': self.episode_user_ratings,
                'movie': self.movie_user_ratings,
            },
        }

    @staticmethod
    def lookup_series(session, title=None, year=None, only_cached=None, **lookup_params):
        trakt_show_ids = db.TraktShowIds(**lookup_params)
        series = db.get_item_from_cache(
            db.TraktShow, title=title, year=year, trakt_ids=trakt_show_ids, session=session
        )
        found = None
        if not series and title:
            found = (
                session.query(db.TraktShowSearchResult)
                .filter(db.TraktShowSearchResult.search == title.lower())
                .first()
            )
            if found and found.series:
                logger.debug(
                    'Found {} in previous search results as {}', title, found.series.title
                )
                series = found.series
        if only_cached:
            if series:
                return series
            raise LookupError('Series %s not found from cache' % lookup_params)
        if series and not series.expired:
            return series
        try:
            trakt_show = db.get_trakt_data(
                'show', title=title, year=year, trakt_ids=trakt_show_ids
            )
        except LookupError as e:
            if series:
                logger.debug('Error refreshing show data from trakt, using cached. {}', e)
                return series
            raise
        try:
            series = session.merge(db.TraktShow(trakt_show, session))
            if series and title.lower() == series.title.lower():
                return series
            elif series and title and not found:
                if (
                    not session.query(db.TraktShowSearchResult)
                    .filter(db.TraktShowSearchResult.search == title.lower())
                    .first()
                ):
                    logger.debug('Adding search result to db')
                    session.merge(db.TraktShowSearchResult(search=title, series=series))
            elif series and found:
                logger.debug('Updating search result in db')
                found.series = series
            return series
        finally:
            session.commit()

    @staticmethod
    def lookup_movie(session, title=None, year=None, only_cached=None, **lookup_params):
        trakt_movie_ids = db.TraktMovieIds(**lookup_params)
        movie = db.get_item_from_cache(
            db.TraktMovie, title=title, year=year, trakt_ids=trakt_movie_ids, session=session
        )
        found = None
        if not movie and title:
            found = (
                session.query(db.TraktMovieSearchResult)
                .filter(db.TraktMovieSearchResult.search == title.lower())
                .first()
            )
            if found and found.movie:
                logger.debug('Found {} in previous search results as {}', title, found.movie.title)
                movie = found.movie
        if only_cached:
            if movie:
                return movie
            raise LookupError('Movie %s not found from cache' % lookup_params)
        if movie and not movie.expired:
            return movie
        # Parse the movie for better results
        parsed_title = None
        y = year
        if title:
            title_parser = plugin.get('parsing', 'api_trakt').parse_movie(title)
            y = year or title_parser.year
            parsed_title = title_parser.name
        try:
            trakt_movie = db.get_trakt_data(
                'movie', title=parsed_title, year=y, trakt_ids=trakt_movie_ids
            )
        except LookupError as e:
            if movie:
                logger.debug('Error refreshing movie data from trakt, using cached. {}', e)
                return movie
            raise
        try:
            movie = session.merge(db.TraktMovie(trakt_movie, session))
            if movie and title.lower() == movie.title.lower():
                return movie
            if movie and title and not found:
                if (
                    not session.query(db.TraktMovieSearchResult)
                    .filter(db.TraktMovieSearchResult.search == title.lower())
                    .first()
                ):
                    logger.debug('Adding search result to db')
                    session.merge(db.TraktMovieSearchResult(search=title, movie=movie))
            elif movie and found:
                logger.debug('Updating search result in db')
                found.movie = movie
            return movie
        finally:
            session.commit()

    def is_show_in_collection(self, trakt_data, title):
        cache = user_cache.get_shows_collection(self.username, account=self.account)

        in_collection = False
        if trakt_data.id in cache:
            series = cache[trakt_data.id]
            # specials are not included
            number_of_collected_episodes = sum(
                len(s['episodes']) for s in series['seasons'] if s['number'] > 0
            )
            in_collection = number_of_collected_episodes >= trakt_data.aired_episodes

        logger.debug(
            'The result for show entry "{}" is: {}',
            title,
            'Owned' if in_collection else 'Not owned',
        )
        return in_collection

    def is_season_in_collection(self, trakt_data, title):
        cache = user_cache.get_shows_collection(
            username=db.get_username(self.username, self.account), account=self.account
        )

        in_collection = False
        if trakt_data.show.id in cache:
            series = cache[trakt_data.show.id]
            for s in series['seasons']:
                if trakt_data.number == s['number']:
                    in_collection = True
                    break

        logger.debug(
            'The result for season entry "{}" is: {}',
            title,
            'Owned' if in_collection else 'Not owned',
        )
        return in_collection

    def is_episode_in_collection(self, trakt_data, title):
        cache = user_cache.get_shows_collection(self.username, self.account)

        in_collection = False
        if trakt_data.show.id in cache:
            series = cache[trakt_data.show.id]
            for s in series['seasons']:
                if s['number'] == trakt_data.season:
                    # extract all episode numbers currently in collection for the season number
                    episodes = [ep['number'] for ep in s['episodes']]
                    in_collection = trakt_data.number in episodes
                    break

        logger.debug(
            'The result for episode entry "{}" is: {}',
            title,
            'Owned' if in_collection else 'Not owned',
        )
        return in_collection

    def is_movie_in_collection(self, trakt_data, title):
        cache = user_cache.get_movie_collection(self.username, self.account)
        in_collection = trakt_data.id in cache
        logger.debug(
            'The result for movie entry "{}" is: {}',
            title,
            'Owned' if in_collection else 'Not owned',
        )
        return in_collection

    def is_show_watched(self, trakt_data, title):
        cache = user_cache.get_watched_shows(username=self.username, account=self.account)
        is_watched = False
        if trakt_data.id in cache:
            series = cache[trakt_data.id]
            # specials are not included
            number_of_watched_episodes = sum(
                len(s['episodes']) for s in series['seasons'] if s['number'] > 0
            )
            is_watched = number_of_watched_episodes == trakt_data.aired_episodes

        logger.debug(
            'The result for show entry "{}" is: {}',
            title,
            'Watched' if is_watched else 'Not watched',
        )
        return is_watched

    def is_season_watched(self, trakt_data, title):
        cache = user_cache.get_watched_shows(username=self.username, account=self.account)

        is_watched = False
        if trakt_data.show.id in cache:
            series = cache[trakt_data.show.id]
            for s in series['seasons']:
                if trakt_data.number == s['number']:
                    is_watched = True
                    break
        logger.debug(
            'The result for season entry "{}" is: {}',
            title,
            'Watched' if is_watched else 'Not watched',
        )
        return is_watched

    def is_episode_watched(self, trakt_data, title):
        cache = user_cache.get_watched_shows(username=self.username, account=self.account)
        is_watched = False
        if trakt_data.show.id in cache:
            series = cache[trakt_data.show.id]
            for s in series['seasons']:
                if s['number'] == trakt_data.season:
                    # extract all episode numbers currently in collection for the season number
                    episodes = [ep['number'] for ep in s['episodes']]
                    is_watched = trakt_data.number in episodes
                    break
        logger.debug(
            'The result for episode entry "{}" is: {}',
            title,
            'Watched' if is_watched else 'Not watched',
        )
        return is_watched

    def is_movie_watched(self, trakt_data, title):
        cache = user_cache.get_watched_movies(username=self.username, account=self.account)
        is_watched = trakt_data.id in cache
        logger.debug(
            'The result for movie entry "{}" is: {}',
            title,
            'Watched' if is_watched else 'Not watched',
        )
        return is_watched

    def show_user_ratings(self, trakt_data, title):
        cache = user_cache.get_show_user_ratings(username=self.username, account=self.account)
        user_rating = None
        if trakt_data.id in cache:
            user_rating = cache[trakt_data.id]['rating']
        logger.debug('User rating for show entry "{}" is: {}', title, user_rating)
        return user_rating

    def season_user_ratings(self, trakt_data, title):
        cache = user_cache.get_season_user_ratings(username=self.username, account=self.account)
        user_rating = None
        if trakt_data.id in cache and trakt_data.number == cache[trakt_data.id]['number']:
            user_rating = cache[trakt_data.id]['rating']
        logger.debug('User rating for season entry "{}" is: {}', title, user_rating)
        return user_rating

    def episode_user_ratings(self, trakt_data, title):
        cache = user_cache.get_episode_user_ratings(username=self.username, account=self.account)
        user_rating = None
        if trakt_data.id in cache:
            user_rating = cache[trakt_data.id]['rating']
        logger.debug('User rating for episode entry "{}" is: {}', title, user_rating)
        return user_rating

    def movie_user_ratings(self, trakt_data, title):
        cache = user_cache.get_movie_user_ratings(username=self.username, account=self.account)
        user_rating = None
        if trakt_data.id in cache:
            user_rating = cache[trakt_data.id]['rating']
        logger.debug('User rating for movie entry "{}" is: {}', title, user_rating)
        return user_rating


@event('plugin.register')
def register_plugin():
    plugin.register(ApiTrakt, 'api_trakt', api_ver=2, interfaces=[])



import os

import yaml
from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more, process_config
from flexget.event import event
from flexget.utils.tools import MergeException

plugin_name = 'include'
logger = logger.bind(name=plugin_name)


class PluginInclude:
    """
    Include configuration from another yaml file.

    Example::

      include: series.yml

    File content must be valid for a task configuration
    """

    schema = one_or_more({'type': 'string'})

    @plugin.priority(256)
    def on_task_prepare(self, task, config):
        if not config:
            return

        files = config
        if isinstance(config, str):
            files = [config]

        for file_name in files:
            file = os.path.expanduser(file_name)
            if not os.path.isabs(file):
                file = os.path.join(task.manager.config_base, file)
            with open(file, encoding='utf-8') as inc_file:
                include = yaml.safe_load(inc_file)
                inc_file.flush()
            errors = process_config(include, plugin.plugin_schemas(interface='task'))
            if errors:
                logger.error('Included file {} has invalid config:', file)
                for error in errors:
                    logger.error('[{}] {}', error.json_pointer, error.message)
                task.abort('Invalid config in included file %s' % file)

            logger.debug('Merging {} into task {}', file, task.name)
            # merge
            try:
                task.merge_config(include)
            except MergeException:
                raise plugin.PluginError(
                    'Failed to merge include file to task %s, incompatible datatypes' % task.name
                )


@event('plugin.register')
def register_plugin():
    plugin.register(PluginInclude, 'include', api_ver=2, builtin=True)



import contextlib
import sys
import threading
from textwrap import wrap
from typing import Iterator, List, Optional, TextIO, Union

from colorclass import Color, Windows
from terminaltables import (
    AsciiTable,
    DoubleTable,
    GithubFlavoredMarkdownTable,
    PorcelainTable,
    SingleTable,
)
from terminaltables.base_table import BaseTable
from terminaltables.terminal_io import terminal_size

from flexget.options import ArgumentParser
from flexget.utils.tools import io_encoding

# Enable terminal colors on windows.
# pythonw (flexget-headless) does not have a sys.stdout, this command would crash in that case
if sys.platform == 'win32' and sys.stdout:
    Windows.enable(auto_colors=True)


local_context = threading.local()


def terminal_info() -> dict:
    """
    Returns info we need about the output terminal.
    When called over IPC, this function is monkeypatched to return the info about the client terminal.
    """
    return {'size': terminal_size(), 'isatty': sys.stdout.isatty()}


class TerminalTable:
    """
    A data table suited for CLI output, created via its sent parameters. For example::

        header = ['Col1', 'Col2']
        table_data = [header]
        for item in iterable:
            table_data.append([item.attribute1, item.attribute2])
        table = TerminalTable('plain', table_data)
        print table.output

    Optional values are setting table title, and supplying wrap_columns list and
    drop_column list. If table does not fit into terminal any columns listed in
    wrap_columns will be tried to wrap and if resulting columns are below MIN_WIDTH(10)
    columns listed in drop_column will be removed from output.

    Example::

        header = ['Col1', 'Col2']
        table_data = [header]
        for item in iterable:
            table_data.append([item.attribute1, item.attribute2])
        table = TerminalTable('plain', table_data, 'Table title', wrap_columns=[1,2],
                              drop_columns=[4,2])
        print table.output

    :param table_type: A string matching supported_table_types() keys.
    :param table_data: Table data as a list of lists of strings. See `terminaltables` doc.
    :param title: Optional title for table
    :param wrap_columns: A list of column numbers which will can be wrapped.
        In case of multiple values even split is used.
    :param drop_columns: A list of column numbers which can be dropped if needed.
        List in order of priority.
    """

    MIN_WIDTH = 10
    ASCII_TYPES = ['plain', 'porcelain']

    def __init__(
        self,
        table_type: str,
        table_data: List[List[str]],
        title: Optional[str] = None,
        wrap_columns: Optional[List[int]] = None,
        drop_columns: Optional[List[int]] = None,
    ) -> None:
        self.title = title
        self.wrap_columns = wrap_columns or []
        self.drop_columns = drop_columns or []
        self.table_data = table_data

        # Force table type to be ASCII when not TTY and type isn't already ASCII
        if table_type not in self.ASCII_TYPES and not terminal_info()['isatty']:
            self.type = 'porcelain'
        else:
            self.type = table_type

        self._init_table()

    def _init_table(self) -> None:
        """Assigns self.table with the built table based on data."""
        self.table = self._build_table(self.type, self.table_data)
        if self.type == 'porcelain':
            self.table.padding_left = 0
            self.table.padding_right = 0
            return
        adjustable = bool(self.wrap_columns + self.drop_columns)
        if not self.valid_table and adjustable:
            self.table = self._resize_table()

    def _build_table(self, table_type: str, table_data: List[List[str]]) -> BaseTable:
        return self.supported_table_types()[table_type](table_data)

    @property
    def output(self) -> str:
        self.table.title = self.title
        if self.type == 'porcelain':
            return '\n'.join(line.rstrip() for line in self.table.table.splitlines())
        else:
            return '\n' + self.table.table

    @staticmethod
    def supported_table_types() -> dict:
        """This method hold the dict for supported table type."""
        return {
            'plain': AsciiTable,
            'porcelain': PorcelainTable,
            'single': SingleTable,
            'double': DoubleTable,
            'github': GithubFlavoredMarkdownTable,
        }

    @property
    def _columns(self) -> int:
        if not self.table_data:
            return 0
        else:
            return len(self.table_data[0])

    def _longest_rows(self) -> dict:
        """
        Calculate longest line for each column.

        :returns: dictionary where key is column number and value longest value
        """
        longest = {c: 0 for c in range(self._columns)}
        for row in self.table_data:
            for index, value in enumerate(row):
                if len(str(value)) > longest[index]:
                    longest[index] = len(str(value))
        return longest

    @property
    def valid_table(self):
        return self.table.table_width <= terminal_info()['size'][0]

    def _calc_wrap(self) -> Optional[int]:
        """
        :return: Calculated wrap value to be used for self.wrap_columns.
            If wraps are not defined None is returned
        """
        if not self.wrap_columns:
            return None
        longest = self._longest_rows()
        margin = self._columns * 2 + self._columns + 1
        static_columns = sum(longest.values())
        for wrap_c in self.wrap_columns:
            static_columns -= longest[wrap_c]
        space_left = terminal_info()['size'][0] - static_columns - margin
        # TODO: This is a bit dumb if wrapped columns have huge disparity
        # for example in flexget plugins the phases and flags
        return int(space_left / len(self.wrap_columns))

    def _drop_column(self, col: int) -> None:
        name = self.table_data[0][col]
        console('Least important column `%s` removed due terminal size' % name)

        for row in self.table_data:
            del row[col]

        def adjust(l):
            new_vals = []
            for c in l:
                if c == col:
                    # column removed
                    continue
                if c > col:
                    # column before removed
                    new_vals.append(c - 1)
                else:
                    new_vals.append(c)
            return new_vals

        self.wrap_columns = adjust(self.wrap_columns)
        self.drop_columns = adjust(self.drop_columns)

    def _drop_columns_with_wrap(self) -> None:
        """Drop columns until wrapped columns fit or are removed as well"""
        while self.drop_columns and self.wrap_columns:
            drop = self.drop_columns.pop(0)
            self._drop_column(drop)
            wrapped_width = self._calc_wrap()
            if wrapped_width and wrapped_width < self.MIN_WIDTH:
                continue
            else:
                return

    def _resize_table(self) -> BaseTable:
        """Adjust table data and columns to fit into terminal"""
        if self.drop_columns and not self.wrap_columns:
            while self.drop_columns and not self.valid_table:
                drop = self.drop_columns.pop(0)
                self._drop_column(drop)

        wrapped_width = self._calc_wrap()
        if wrapped_width and wrapped_width < self.MIN_WIDTH:
            self._drop_columns_with_wrap()
            wrapped_width = self._calc_wrap()

        # construct new table
        output_table = []
        for row in self.table_data:
            output_row = []
            for col_num, value in enumerate(row):
                output_value = value
                if col_num in self.wrap_columns:
                    # This probably shouldn't happen, can be caused by wrong parameters sent to drop_columns and
                    # wrap_columns
                    if wrapped_width and wrapped_width <= 0:
                        raise TerminalTableError(
                            'Table could not be rendered correctly using it given data'
                        )
                    output_value = word_wrap(str(value), wrapped_width)
                output_row.append(output_value)
            output_table.append(output_row)
        return self._build_table(self.type, output_table)


class TerminalTableError(Exception):
    """ A CLI table error"""


table_parser = ArgumentParser(add_help=False)
table_parser.add_argument(
    '--table-type',
    choices=list(TerminalTable.supported_table_types()),
    default='single',
    help='Select output table style',
)
table_parser.add_argument(
    '--porcelain',
    dest='table_type',
    action='store_const',
    const='porcelain',
    help='Make the output parseable. Similar to using `--table-type porcelain`',
)


def word_wrap(text: str, max_length: int) -> str:
    """A helper method designed to return a wrapped string.

    :param text: Text to wrap
    :param max_length: Maximum allowed string length
    :return: Wrapped text or original text
    """
    if len(text) >= max_length:
        return '\n'.join(wrap(text, max_length))
    return text


def colorize(color: str, text: str, auto: bool = True) -> Union[Color, str]:
    """
    A simple override of Color.colorize which sets the default auto colors value to True, since it's the more common
    use case. When output isn't TTY just return text

    :param color: Color tag to use
    :param text: Text to color
    :param auto: Whether to apply auto colors

    :return: Colored text or text
    """
    if not terminal_info()['isatty']:
        return text
    return Color.colorize(color, text, auto)


@contextlib.contextmanager
def capture_console(filelike: TextIO) -> Iterator:
    old_output = get_console_output()
    local_context.output = filelike
    try:
        yield
    finally:
        local_context.output = old_output


def get_console_output() -> Optional[TextIO]:
    return getattr(local_context, 'output', None)


def console(text: str, *args, **kwargs) -> None:
    """
    Print to console safely. Output is able to be captured by different streams in different contexts.

    Any plugin wishing to output to the user's console should use this function instead of print so that
    output can be redirected when FlexGet is invoked from another process.

    Accepts arguments like the `print` function does.
    """
    kwargs['file'] = getattr(local_context, 'output', sys.stdout)
    try:
        print(text, *args, **kwargs)
    except UnicodeEncodeError:
        text = text.encode(io_encoding, 'replace').decode(io_encoding)
        print(text, *args, **kwargs)
    kwargs['file'].flush()  # flush to make sure the output is printed right away



import sys
from pathlib import Path

import flexget
from flexget import options
from flexget.event import event
from flexget.terminal import console
from flexget.utils.tools import io_encoding


def print_debug_info(manager, options):
    install_location = Path(__file__).absolute().parent.parent.parent
    console('FlexGet Version: {}'.format(flexget.__version__))
    console('Install location: {}'.format(install_location))
    console('Config location: {}'.format(manager.config_path))
    console('Python version: {}.{}.{}'.format(*sys.version_info[:3]))
    console('Detected IO encoding: {}'.format(io_encoding))


@event('options.register')
def register_parser_arguments():
    options.register_command(
        'debug-info', print_debug_info, help='display useful info for debugging'
    )



import re

from loguru import logger
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from flexget import plugin
from flexget.event import event

PLUGIN_NAME = 'download_auth'

logger = logger.bind(name=PLUGIN_NAME)


class DownloadAuth:
    host_schema = {
        'additionalProperties': {
            'type': 'object',
            'properties': {
                'username': {'type': 'string'},
                'password': {'type': 'string'},
                'type': {'type': 'string', 'enum': ['basic', 'digest'], 'default': 'basic'},
            },
            'required': ['username', 'password'],
        }
    }
    schema = {'type': 'array', 'items': host_schema, 'minimumItems': 1}

    auth_mapper = {'basic': HTTPBasicAuth, 'digest': HTTPDigestAuth}

    # Run before all downloads
    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_download(self, task, config):
        for entry in task.accepted:
            if entry.get('download_auth'):
                logger.debug('entry {} already has auth set, skipping', entry)
                continue
            for host_config in config:
                for host, auth_config in host_config.items():
                    if re.search(host, entry['url'], re.IGNORECASE):
                        auth_type = auth_config['type']
                        username = auth_config['username']
                        password = auth_config['password']
                        logger.debug('setting auth type {} with username {}', auth_type, username)
                        entry['download_auth'] = self.auth_mapper[auth_type](username, password)


@event('plugin.register')
def register_plugin():
    plugin.register(DownloadAuth, PLUGIN_NAME, api_ver=2)



import os
import re
import socket
from io import BytesIO
from time import sleep
from urllib.parse import urljoin, urlparse, urlsplit
from xmlrpc import client as xmlrpc_client

from loguru import logger
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.bittorrent import Torrent, is_torrent_file
from flexget.utils.pathscrub import pathscrub
from flexget.utils.template import RenderError

logger = logger.bind(name='rtorrent')


class _Method:
    # some magic to bind an XML-RPC method to an RPC server.
    # supports "nested" methods (e.g. examples.getStateName)

    def __init__(self, send, name):
        self.__send = send
        self.__name = name

    def __getattr__(self, name):
        return _Method(self.__send, "%s.%s" % (self.__name, name))

    def __call__(self, *args):
        return self.__send(self.__name, args)


class HTTPDigestTransport(xmlrpc_client.Transport):
    """
    Transport that uses requests to support Digest authentication.
    """

    def __init__(self, scheme, digest_auth, username, password, session, *args, **kwargs):
        self.__scheme = scheme
        self.__session = session
        self.__digest_auth = digest_auth
        self.__username = username
        self.__password = password
        self.verbose = 0
        xmlrpc_client.Transport.__init__(self, *args, **kwargs)  # old style class

    def request(self, host, handler, request_body, verbose=False):
        return self.single_request(host, handler, request_body, verbose)

    def single_request(self, host, handler, request_body, verbose=0):
        url = urljoin('{0}://{1}'.format(self.__scheme, host), handler)

        auth = self.get_auth()
        response = self.send_request(url, auth, request_body)

        # if status code is 401, it means we used the wrong auth method
        if response.status_code == 401:
            logger.warning(
                '{} auth failed. Retrying with {}. Please change your config.',
                'Digest' if self.__digest_auth else 'Basic',
                'Basic' if self.__digest_auth else 'Digest',
            )
            self.__digest_auth = not self.__digest_auth

            auth = self.get_auth()
            response = self.send_request(url, auth, request_body)

        response.raise_for_status()

        return self.parse_response(response)

    def get_auth(self):
        if self.__digest_auth:
            return HTTPDigestAuth(self.__username, self.__password)
        return HTTPBasicAuth(self.__username, self.__password)

    def send_request(self, url, auth, data):
        return self.__session.post(url, auth=auth, data=data, raise_status=False)

    def parse_response(self, response):
        p, u = self.getparser()

        if self.verbose:
            logger.info('body: {!r}', response)

        p.feed(response.content)
        p.close()

        return u.close()


def encode_netstring(input):
    return str(len(input)).encode() + b':' + input + b','


def encode_header(key, value):
    return key + b'\x00' + value + b'\x00'


class SCGITransport(xmlrpc_client.Transport):
    """
    Public domain SCGITrannsport implementation from:
    https://github.com/JohnDoee/autotorrent/blob/develop/autotorrent/scgitransport.py
    """

    def __init__(self, *args, **kwargs):
        self.socket_path = kwargs.pop('socket_path', '')
        xmlrpc_client.Transport.__init__(self, *args, **kwargs)

    def single_request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        if self.socket_path:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect(self.socket_path)
        else:
            host, port = host.split(':')
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, int(port)))

        request = encode_header(b'CONTENT_LENGTH', str(len(request_body)).encode())
        request += encode_header(b'SCGI', b'1')
        request += encode_header(b'REQUEST_METHOD', b'POST')
        request += encode_header(b'REQUEST_URI', handler.encode())

        request = encode_netstring(request)
        request += request_body

        s.send(request)

        response = b''
        while True:
            r = s.recv(1024)
            if not r:
                break
            response += r

        response_body = BytesIO(b'\r\n\r\n'.join(response.split(b'\r\n\r\n')[1:]))

        return self.parse_response(response_body)


if not hasattr(xmlrpc_client.Transport, 'single_request'):
    SCGITransport.request = SCGITransport.single_request


def create_proxy(url):
    parsed = urlsplit(url)
    if not parsed.scheme:
        path = parsed.path
        return xmlrpc_client.ServerProxy('http://1', transport=SCGITransport(socket_path=path))
    if parsed.scheme == 'scgi':
        url = 'http://%s' % parsed.netloc
        return xmlrpc_client.ServerProxy(url, transport=SCGITransport())
    logger.debug('Creating Normal XMLRPC Proxy with url {!r}', url)
    return xmlrpc_client.ServerProxy(url)


class RTorrent:
    """ rTorrent API client """

    default_fields = (
        'hash',
        'name',
        'up_total',
        'down_total',
        'down_rate',
        'is_open',
        'is_active',
        'custom1',
        'custom2',
        'custom3',
        'custom4',
        'custom5',
        'state',
        'complete',
        'bytes_done',
        'down.rate',
        'left_bytes',
        'ratio',
        'base_path',
        'load_date',
    )

    required_fields = ('hash', 'name', 'base_path')

    def __init__(self, uri, username=None, password=None, digest_auth=None, session=None):
        """
        New connection to rTorrent

        :param uri: RTorrent URL. Supports both http(s) and scgi
        :param username: Username for basic auth over http(s)
        :param password: Password for basic auth over http(s)
        """
        self.uri = uri
        self.username = username
        self.password = password
        self.digest_auth = digest_auth
        self._version = None

        parsed_uri = urlparse(uri)

        if self.username and self.password and parsed_uri.scheme not in ['http', 'https']:
            raise OSError('Username and password only supported on http(s)')

        # Determine the proxy server
        if parsed_uri.scheme in ['http', 'https']:
            sp = xmlrpc_client.ServerProxy
        elif parsed_uri.scheme == 'scgi':
            sp = create_proxy
        elif parsed_uri.scheme == '' and parsed_uri.path:
            self.uri = parsed_uri.path
            sp = create_proxy
        else:
            raise OSError('Unsupported scheme %s for uri %s' % (parsed_uri.scheme, self.uri))

        # Use a special transport if http(s)
        if parsed_uri.scheme in ['http', 'https']:
            self._server = sp(
                self.uri,
                transport=HTTPDigestTransport(
                    parsed_uri.scheme, self.digest_auth, self.username, self.password, session
                ),
            )
        else:
            self._server = sp(self.uri)

    def _clean_fields(self, fields, reverse=False):
        if not fields:
            fields = list(self.default_fields)

        if reverse:
            for field in ['up.total', 'down.total', 'down.rate']:
                if field in fields:
                    fields[fields.index(field)] = field.replace('.', '_')
            return fields

        for required_field in self.required_fields:
            if required_field not in fields:
                fields.insert(0, required_field)

        for field in ['up_total', 'down_total', 'down_rate']:
            if field in fields:
                fields[fields.index(field)] = field.replace('_', '.')

        return fields

    def load(self, raw_torrent, fields=None, start=False, mkdir=True):

        if fields is None:
            fields = {}
        # First param is empty 'target'
        params = ['', xmlrpc_client.Binary(raw_torrent)]

        # Additional fields to set
        for key, val in fields.items():
            # Values must be escaped if within params
            # TODO: What are the escaping requirements? re.escape works differently on python 3.7+
            params.append('d.%s.set=%s' % (key, re.escape(str(val))))

        if mkdir and 'directory' in fields:
            result = self._server.execute.throw('', 'mkdir', '-p', fields['directory'])
            if result != 0:
                raise xmlrpc_client.Error('Failed creating directory %s' % fields['directory'])

        # by default rtorrent won't allow calls over 512kb in size.
        xmlrpc_size = (
            len(xmlrpc_client.dumps(tuple(params), 'raw_start')) + 71680
        )  # Add 70kb for buffer
        if xmlrpc_size > 524288:
            prev_size = self._server.network.xmlrpc.size_limit()
            self._server.network.xmlrpc.size_limit.set('', xmlrpc_size)

        # Call load method and return the response
        if start:
            result = self._server.load.raw_start(*params)
        else:
            result = self._server.load.raw(*params)

        if xmlrpc_size > 524288:
            self._server.network.xmlrpc.size_limit.set('', prev_size)

        return result

    def get_directory(self):
        return self._server.get_directory()

    def torrent(self, info_hash, fields=None):
        """ Get the details of a torrent """
        if not fields:
            fields = list(self.default_fields)

        fields = self._clean_fields(fields)

        multi_call = xmlrpc_client.MultiCall(self._server)

        for field in fields:
            method_name = 'd.%s' % field
            getattr(multi_call, method_name)(info_hash)

        resp = multi_call()
        # TODO: Maybe we should return a named tuple or a Torrent class?
        return dict(list(zip(self._clean_fields(fields, reverse=True), [val for val in resp])))

    def torrents(self, view='main', fields=None):
        if not fields:
            fields = list(self.default_fields)
        fields = self._clean_fields(fields)

        params = ['d.%s=' % field for field in fields]
        params.insert(0, view)

        resp = self._server.d.multicall2('', params)

        # Response is formatted as a list of lists, with just the values
        return [dict(list(zip(self._clean_fields(fields, reverse=True), val))) for val in resp]

    def update(self, info_hash, fields):
        multi_call = xmlrpc_client.MultiCall(self._server)

        for key, val in fields.items():
            method_name = 'd.%s.set' % key
            getattr(multi_call, method_name)(info_hash, val)

        return multi_call()[0]

    def delete(self, info_hash):
        return self._server.d.erase(info_hash)

    def stop(self, info_hash):
        self._server.d.stop(info_hash)
        return self._server.d.close(info_hash)

    def start(self, info_hash):
        return self._server.d.start(info_hash)

    def move(self, info_hash, dst_path):
        self.stop(info_hash)

        torrent = self.torrent(info_hash, fields=['base_path'])

        try:
            logger.verbose('Creating destination directory `{}`', dst_path)
            self._server.execute.throw('', 'mkdir', '-p', dst_path)
        except xmlrpc_client.Error:
            raise xmlrpc_client.Error("unable to create folder %s" % dst_path)

        self._server.execute.throw('', 'mv', '-u', torrent['base_path'], dst_path)
        self._server.d.set_directory(info_hash, dst_path)
        self.start(info_hash)


class RTorrentPluginBase:
    priority_map = {'high': 3, 'medium': 2, 'low': 1, 'off': 0}

    def _build_options(self, config, entry, entry_first=True):
        options = {}

        for opt_key in (
            'path',
            'message',
            'priority',
            'custom1',
            'custom2',
            'custom3',
            'custom4',
            'custom5',
        ):
            # Values do not merge config with task
            # Task takes priority then config is used
            entry_value = entry.get(opt_key)
            config_value = config.get(opt_key)

            if entry_first:
                if entry_value:
                    options[opt_key] = entry.render(entry_value)
                elif config_value:
                    options[opt_key] = entry.render(config_value)
            else:
                if config_value:
                    options[opt_key] = entry.render(config_value)
                elif entry_value:
                    options[opt_key] = entry.render(entry_value)

        # Convert priority from string to int
        priority = options.get('priority')
        if priority and priority in self.priority_map:
            options['priority'] = self.priority_map[priority]

        # Map Flexget path to directory in rTorrent
        if options.get('path'):
            options['directory'] = options['path']
            del options['path']

        if 'directory' in options:
            options['directory'] = pathscrub(options['directory'])

        return options


class RTorrentOutputPlugin(RTorrentPluginBase):
    schema = {
        'type': 'object',
        'properties': {
            # connection info
            'uri': {'type': 'string'},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'digest_auth': {'type': 'boolean', 'default': False},
            'start': {'type': 'boolean', 'default': True},
            'mkdir': {'type': 'boolean', 'default': True},
            'action': {'type': 'string', 'emun': ['update', 'delete', 'add'], 'default': 'add'},
            # properties to set on rtorrent download object
            'message': {'type': 'string'},
            'priority': {'type': 'string'},
            'path': {'type': 'string'},
            'custom1': {'type': 'string'},
            'custom2': {'type': 'string'},
            'custom3': {'type': 'string'},
            'custom4': {'type': 'string'},
            'custom5': {'type': 'string'},
            'fast_resume': {'type': 'boolean', 'default': False},
        },
        'required': ['uri'],
        'additionalProperties': False,
    }

    def _verify_load(self, client, info_hash):
        ex = xmlrpc_client.Error()
        for _ in range(0, 5):
            try:
                return client.torrent(info_hash, fields=['hash'])
            except xmlrpc_client.Error as e:
                ex = e
                sleep(0.5)
        raise ex

    @plugin.priority(120)
    def on_task_download(self, task, config):
        # If the download plugin is not enabled, we need to call it to get
        # our temp .torrent files
        if config['action'] == 'add' and 'download' not in task.config:
            download = plugin.get('download', self)
            download.get_temp_files(task, handle_magnets=True, fail_html=True)

    @plugin.priority(135)
    def on_task_output(self, task, config):

        client = RTorrent(
            os.path.expanduser(config['uri']),
            username=config.get('username'),
            password=config.get('password'),
            digest_auth=config['digest_auth'],
            session=task.requests,
        )

        try:
            for entry in task.accepted:
                if config['action'] == 'add':
                    if task.options.test:
                        logger.info('Would add {} to rTorrent', entry['url'])
                        continue
                    try:
                        options = self._build_options(config, entry)
                    except RenderError as e:
                        entry.fail("failed to render properties %s" % str(e))
                        continue

                    # fast_resume is not really an rtorrent option so it's not in _build_options
                    fast_resume = entry.get('fast_resume', config['fast_resume'])
                    self.add_entry(
                        client,
                        entry,
                        options,
                        start=config['start'],
                        mkdir=config['mkdir'],
                        fast_resume=fast_resume,
                    )

                info_hash = entry.get('torrent_info_hash')

                if not info_hash:
                    entry.fail('Failed to %s as no info_hash found' % config['action'])
                    continue

                if config['action'] == 'delete':
                    if task.options.test:
                        logger.info(
                            'Would delete {} ({}) from rTorrent',
                            entry['title'],
                            entry['torrent_info_hash'],
                        )
                        continue
                    self.delete_entry(client, entry)

                if config['action'] == 'update':
                    if task.options.test:
                        logger.info(
                            'Would update {} ({}) in rTorrent',
                            entry['title'],
                            entry['torrent_info_hash'],
                        )
                        continue
                    self.update_entry(client, entry, config)

        except OSError as e:
            raise plugin.PluginError("Couldn't connect to rTorrent: %s" % str(e))

    def delete_entry(self, client, entry):
        try:
            client.delete(entry['torrent_info_hash'])
            logger.verbose(
                'Deleted {} ({}) in rtorrent ', entry['title'], entry['torrent_info_hash']
            )
        except xmlrpc_client.Error as e:
            entry.fail('Failed to delete: %s' % str(e))
            return

    def update_entry(self, client, entry, config):
        info_hash = entry['torrent_info_hash']

        # First check if it already exists
        try:
            existing = client.torrent(info_hash, fields=['base_path'])
        except xmlrpc_client.Error:
            existing = False

        # Build options but make config values override entry values
        try:
            options = self._build_options(config, entry, entry_first=False)
        except RenderError as e:
            entry.fail("failed to render properties %s" % str(e))
            return

        if existing and 'directory' in options:
            # Check if changing to another directory which requires a move
            if options['directory'] != existing['base_path'] and options[
                'directory'
            ] != os.path.dirname(existing['base_path']):
                try:
                    logger.verbose(
                        "Path is changing, moving files from '{}' to '{}'",
                        existing['base_path'],
                        options['directory'],
                    )
                    client.move(info_hash, options['directory'])
                except xmlrpc_client.Error as e:
                    entry.fail('Failed moving torrent: %s' % str(e))
                    return

        # Remove directory from update otherwise rTorrent will append the title to the directory path
        if 'directory' in options:
            del options['directory']

        try:
            client.update(info_hash, options)
            logger.verbose('Updated {} ({}) in rtorrent ', entry['title'], info_hash)
        except xmlrpc_client.Error as e:
            entry.fail('Failed to update: %s' % str(e))
            return

    def add_entry(self, client, entry, options, start=True, mkdir=False, fast_resume=False):

        if 'torrent_info_hash' not in entry:
            entry.fail('missing torrent_info_hash')
            return

        if entry['url'].startswith('magnet:'):
            torrent_raw = 'd10:magnet-uri%d:%se' % (len(entry['url']), entry['url'])
            torrent_raw = torrent_raw.encode('ascii')
        else:
            # Check that file is downloaded
            if 'file' not in entry:
                raise plugin.PluginError('Temporary download file is missing from entry')

            # Verify the temp file exists
            if not os.path.exists(entry['file']):
                raise plugin.PluginError('Temporary download file is missing from disk')

            # Verify valid torrent file
            if not is_torrent_file(entry['file']):
                entry.fail("Downloaded temp file '%s' is not a torrent file" % entry['file'])
                return

            # Modify the torrent with resume data if needed
            if fast_resume:
                base = options.get('directory')
                if not base:
                    base = client.get_directory()

                piece_size = entry['torrent'].piece_size
                chunks = int((entry['torrent'].size + piece_size - 1) / piece_size)
                files = []

                for f in entry['torrent'].get_filelist():
                    relative_file_path = os.path.join(f['path'], f['name'])
                    if entry['torrent'].is_multi_file:
                        relative_file_path = os.path.join(
                            entry['torrent'].name, relative_file_path
                        )
                    file_path = os.path.join(base, relative_file_path)
                    # TODO should it simply add the torrent anyway?
                    if not os.path.exists(file_path) and not os.path.isfile(file_path):
                        entry.fail('%s does not exist. Cannot add fast resume data.' % file_path)
                        return
                    # cannot bencode floats, so we need to coerce to int
                    mtime = int(os.path.getmtime(file_path))
                    # priority 0 should be "don't download"
                    files.append({'priority': 0, 'mtime': mtime})

                entry['torrent'].set_libtorrent_resume(chunks, files)
                # Since we modified the torrent, we need to write it to entry['file'] again
                with open(entry['file'], 'wb+') as f:
                    f.write(entry['torrent'].encode())
            try:
                with open(entry['file'], 'rb') as f:
                    torrent_raw = f.read()
            except OSError as e:
                entry.fail('Failed to add to rTorrent %s' % str(e))
                return

            try:
                Torrent(torrent_raw)
            except SyntaxError as e:
                entry.fail('Strange, unable to decode torrent, raise a BUG: %s' % str(e))
                return

        # First check if it already exists
        try:
            if client.torrent(entry['torrent_info_hash']):
                logger.warning("Torrent {} already exists, won't add", entry['title'])
                return
        except xmlrpc_client.Error:
            # No existing found
            pass

        try:
            resp = client.load(torrent_raw, fields=options, start=start, mkdir=mkdir)
            if resp != 0:
                entry.fail('Failed to add to rTorrent invalid return value %s' % resp)
        except xmlrpc_client.Error as e:
            logger.exception(e)
            entry.fail('Failed to add to rTorrent %s' % str(e))
            return

        # Verify the torrent loaded
        try:
            self._verify_load(client, entry['torrent_info_hash'])
            logger.info('{} added to rtorrent', entry['title'])
        except xmlrpc_client.Error as e:
            logger.warning('Failed to verify torrent {} loaded: {}', entry['title'], str(e))

    def on_task_learn(self, task, config):
        """ Make sure all temp files are cleaned up when entries are learned """
        # If download plugin is enabled, it will handle cleanup.
        if 'download' not in task.config:
            download = plugin.get('download', self)
            download.cleanup_temp_files(task)

    on_task_abort = on_task_learn


class RTorrentInputPlugin(RTorrentPluginBase):
    schema = {
        'type': 'object',
        'properties': {
            'uri': {'type': 'string'},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'digest_auth': {'type': 'boolean', 'default': False},
            'view': {'type': 'string', 'default': 'main'},
            'fields': one_or_more({'type': 'string', 'enum': list(RTorrent.default_fields)}),
        },
        'required': ['uri'],
        'additionalProperties': False,
    }

    def on_task_input(self, task, config):
        client = RTorrent(
            os.path.expanduser(config['uri']),
            username=config.get('username'),
            password=config.get('password'),
            digest_auth=config['digest_auth'],
            session=task.requests,
        )

        fields = config.get('fields')

        try:
            torrents = client.torrents(config['view'], fields=fields)
        except (OSError, xmlrpc_client.Error) as e:
            task.abort('Could not get torrents (%s): %s' % (config['view'], e))
            return

        entries = []

        for torrent in torrents:
            entry = Entry(
                title=torrent['name'],
                url='%s/%s' % (os.path.expanduser(config['uri']), torrent['hash']),
                path=torrent['base_path'],
                torrent_info_hash=torrent['hash'],
            )

            for attr, value in torrent.items():
                entry[attr] = value

            entries.append(entry)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(RTorrentOutputPlugin, 'rtorrent', api_ver=2)
    plugin.register(RTorrentInputPlugin, 'from_rtorrent', api_ver=2)



from math import ceil

from flask import jsonify, request
from loguru import logger
from sqlalchemy.orm.exc import NoResultFound

from flexget.api import APIResource, api
from flexget.api.app import (
    NotFoundError,
    base_message_schema,
    etag,
    pagination_headers,
    success_response,
)

from . import db

logger = logger.bind(name='failed_api')

retry_failed_api = api.namespace('failed', description='View and manage failed entries')


class ObjectsContainer:
    retry_failed_entry_object = {
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
            'title': {'type': 'string'},
            'url': {'type': 'string'},
            'added_at': {'type': 'string', 'format': 'date-time'},
            'reason': {'type': 'string'},
            'count': {'type': 'integer'},
            'retry_time': {'type': ['string', 'null'], 'format': 'date-time'},
        },
        'required': ['id', 'title', 'url', 'added_at', 'reason', 'count', 'retry_time'],
        'additionalProperties': False,
    }
    retry_entries_list_object = {'type': 'array', 'items': retry_failed_entry_object}


retry_failed_entry_schema = api.schema_model(
    'retry_failed_entry_schema', ObjectsContainer.retry_failed_entry_object
)
retry_entries_list_schema = api.schema_model(
    'retry_entries_list_schema', ObjectsContainer.retry_entries_list_object
)

sort_choices = ('failure_time', 'id', 'title', 'url', 'reason', 'count', 'retry_time')
failed_parser = api.pagination_parser(sort_choices=sort_choices)


@retry_failed_api.route('/')
class RetryFailed(APIResource):
    @etag
    @api.response(NotFoundError)
    @api.response(200, model=retry_entries_list_schema)
    @api.doc(parser=failed_parser)
    def get(self, session=None):
        """ List all failed entries """
        args = failed_parser.parse_args()

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_by = args['sort_by']
        sort_order = args['order']

        if sort_by == 'failure_time':
            sort_by = 'tof'

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        descending = sort_order == 'desc'

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {
            'start': start,
            'stop': stop,
            'descending': descending,
            'sort_by': sort_by,
            'session': session,
        }

        total_items = db.get_failures(session, count=True)

        if not total_items:
            return jsonify([])

        failed_entries = [failed.to_dict() for failed in db.get_failures(**kwargs)]

        total_pages = int(ceil(total_items / float(per_page)))

        if page > total_pages:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(failed_entries))

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(failed_entries)

        # Add link header to response
        rsp.headers.extend(pagination)

        return rsp

    @api.response(200, 'successfully deleted failed entry', model=base_message_schema)
    def delete(self, session=None):
        """ Clear all failed entries """
        logger.debug('deleting all failed entries')
        deleted = session.query(db.FailedEntry).delete()
        return success_response('successfully deleted %d failed entries' % deleted)


@retry_failed_api.route('/<int:failed_entry_id>/')
@api.response(NotFoundError)
class RetryFailedID(APIResource):
    @etag
    @api.doc(params={'failed_entry_id': 'ID of the failed entry'})
    @api.response(200, model=retry_failed_entry_schema)
    def get(self, failed_entry_id, session=None):
        """ Get failed entry by ID """
        try:
            failed_entry = (
                session.query(db.FailedEntry).filter(db.FailedEntry.id == failed_entry_id).one()
            )
        except NoResultFound:
            raise NotFoundError('could not find entry with ID %i' % failed_entry_id)
        return jsonify(failed_entry.to_dict())

    @api.response(200, 'successfully delete failed entry', model=base_message_schema)
    def delete(self, failed_entry_id, session=None):
        """ Delete failed entry by ID """
        try:
            failed_entry = (
                session.query(db.FailedEntry).filter(db.FailedEntry.id == failed_entry_id).one()
            )
        except NoResultFound:
            raise NotFoundError('could not find entry with ID %i' % failed_entry_id)
        logger.debug('deleting failed entry: "{}"', failed_entry.title)
        session.delete(failed_entry)
        return success_response('successfully delete failed entry %d' % failed_entry_id)



import re
import unicodedata
from datetime import date, datetime, timedelta

from loguru import logger
from requests.exceptions import HTTPError, RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter
from flexget.utils.soup import get_soup

logger = logger.bind(name='search_npo')

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('npostart.nl', '8 seconds'))


class NPOWatchlist:
    """
    Produces entries for every episode on the user's npostart.nl watchlist (Dutch public television).
    Entries can be downloaded using http://arp242.net/code/download-npo

    If 'remove_accepted' is set to 'yes', the plugin will delete accepted entries from the watchlist after download
        is complete.
    If 'max_episode_age_days' is set (and not 0), entries will only be generated for episodes broadcast in the last
        x days.  This only applies to episodes related to series the user is following.
    If 'download_premium' is set to 'yes', the plugin will also download entries that are marked as exclusive
        content for NPO Plus subscribers.

    For example:
        npo_watchlist:
          email: aaaa@bbb.nl
          password: xxx
          remove_accepted: yes
          max_episode_age_days: 7
          download_premium: no
        accept_all: yes
        exec:
          fail_entries: yes
          auto_escape: yes
          on_output:
            for_accepted:
              - download-npo -o "path/to/directory/{{series_name_plain}}" -f "{serie_titel} - {datum} \
                  {aflevering_titel} ({episode_id})" -t {{url}}
        """

    schema = {
        'type': 'object',
        'properties': {
            'email': {'type': 'string'},
            'password': {'type': 'string'},
            'remove_accepted': {'type': 'boolean', 'default': False},
            'max_episode_age_days': {'type': 'integer', 'default': -1},
            'download_premium': {'type': 'boolean', 'default': False},
        },
        'required': ['email', 'password'],
        'additionalProperties': False,
    }

    csrf_token = None

    def _convert_plain(self, s):
        return ''.join(
            c
            for c in unicodedata.normalize('NFD', s)
            if not re.match('Mn|P[^cd]', unicodedata.category(c))
        )

    def _parse_date(self, date_text):
        return datetime.strptime(date_text, '%d-%m-%Y').date()

    def _get_page(self, task, config, url):
        self._login(task, config)
        try:
            logger.debug('Fetching NPO profile page: {}', url)
            page_response = requests.get(url)
            if page_response.url != url:
                raise plugin.PluginError(
                    'Unexpected page: {} (expected {})'.format(page_response.url, url)
                )
            return page_response
        except RequestException as e:
            raise plugin.PluginError('Request error: %s' % str(e))

    def _login(self, task, config):
        if 'isAuthenticatedUser' in requests.cookies:
            logger.debug('Already logged in')
            return

        login_url = 'https://www.npostart.nl/login'
        login_api_url = 'https://www.npostart.nl/api/login'

        try:
            login_response = requests.get(login_url)
            if login_response.url != login_url:
                raise plugin.PluginError('Unexpected login page: {}'.format(login_response.url))

            login_page = get_soup(login_response.content)
            token = login_page.find('input', attrs={'name': '_token'})['value']

            email = config.get('email')
            password = config.get('password')

            profile_response = requests.post(
                login_api_url, {'_token': token, 'username': email, 'password': password}
            )

            if 'isAuthenticatedUser' not in profile_response.cookies:
                raise plugin.PluginError('Failed to login. Check username and password.')
            logger.debug('Succesfully logged in: {}', email)
        except RequestException as e:
            raise plugin.PluginError('Request error: %s' % str(e))

    def _get_favourites_entries(self, task, config, profileId):
        # Details of profile using the $profileId:
        # https://www.npostart.nl/api/account/@me/profile/$profileId  (?refresh=1)
        # XMLHttpRequest could also be done on, but is not reliable, because if a series is
        # bookmarked after episode is already broadcasted, it does not appear in this list.
        # https://www.npostart.nl/ums/accounts/@me/favourites/episodes?dateFrom=2018-06-04&dateTo=2018-06-05
        profile_details_url = 'https://www.npostart.nl/api/account/@me/profile/'

        entries = []
        try:
            profile = requests.get(profile_details_url + profileId).json()
            logger.info('Retrieving favorites for profile {}', profile['name'])
            for favourite in profile['favourites']:
                if favourite['mediaType'] == 'series':
                    entries += self._get_series_episodes(task, config, favourite['mediaId'])
        except RequestException as e:
            raise plugin.PluginError('Request error: %s' % str(e))
        return entries

    def _get_series_episodes(self, task, config, mediaId, series_info=None, page=1):
        episode_tiles_url = 'https://www.npostart.nl/media/series/{0}/episodes'
        episode_tiles_parameters = {
            'page': str(page),
            'tileMapping': 'dedicated',
            'tileType': 'asset',
        }
        entries = []

        if not series_info:
            series_info = self._get_series_info(task, config, mediaId)
            if not series_info:  # if fetching series_info failed, return empty entries
                logger.error('Failed to fetch series information for {}, skipping series', mediaId)
                return entries

        headers = {
            'Origin': 'https://www.npostart.nl',
            'X-XSRF-TOKEN': requests.cookies['XSRF-TOKEN'],
            'X-Requested-With': 'XMLHttpRequest',
        }
        if page > 1:
            headers['Referer'] = episode_tiles_url.format(mediaId) + '?page={0}'.format(
                page - 1
            )  # referer from prev page

        logger.debug(
            'Retrieving episodes page {} for {} ({})', page, series_info['npo_name'], mediaId
        )
        try:
            episodes = requests.get(
                episode_tiles_url.format(mediaId), params=episode_tiles_parameters, headers=headers
            ).json()
            new_entries = self._parse_tiles(task, config, episodes['tiles'], series_info)
            entries += new_entries

            if (
                new_entries and episodes['nextLink']
            ):  # only fetch next page if we accepted any from current page
                logger.debug('NextLink for more episodes: {}', episodes['nextLink'])
                entries += self._get_series_episodes(
                    task,
                    config,
                    mediaId,
                    series_info,
                    page=int(episodes['nextLink'].rsplit('page=')[1]),
                )
        except RequestException as e:
            logger.error('Request error: {}', str(e))  # if it fails, just go to next favourite

        if not entries and page == 1:
            logger.verbose('No new episodes found for {} ({})', series_info['npo_name'], mediaId)
        return entries

    def _get_series_info(self, task, config, mediaId):
        series_info_url = 'https://www.npostart.nl/{0}'
        series_info = None
        logger.verbose('Retrieving series info for {}', mediaId)
        try:
            response = requests.get(series_info_url.format(mediaId))
            logger.debug('Series info found at: {}', response.url)
            page = get_soup(response.content)
            series = page.find('section', class_='npo-header-episode-meta')
            if series:  # sometimes the NPO page does not return valid content
                # create a stub to store the common values for all episodes of this series
                series_info = {
                    'npo_url': response.url,  # we were redirected to the true URL
                    'npo_name': series.find('h1').text,
                    'npo_description': series.find('div', id='metaContent').find('p').text,
                    'npo_language': 'nl',  # hard-code the language as if in NL, for lookup plugins
                    'npo_version': page.find('meta', attrs={'name': 'generator'})['content'],
                }  # include NPO website version
                logger.debug('Parsed series info for: {} ({})', series_info['npo_name'], mediaId)
        except RequestException as e:
            logger.error('Request error: {}', str(e))
        return series_info

    def _parse_tiles(self, task, config, tiles, series_info):
        max_age = config.get('max_episode_age_days')
        download_premium = config.get('download_premium')
        entries = []

        if tiles is not None:
            for tile in tiles:
                # there is only one list_item per tile
                for list_item in get_soup(tile).findAll('div', class_='npo-asset-tile-container'):
                    episode_id = list_item['data-id']
                    premium = 'npo-premium-content' in list_item['class']
                    logger.debug('Parsing episode: {}', episode_id)

                    url = list_item.find('a')['href']
                    # Check if the URL found to the episode matches the expected pattern
                    if len(url.split('/')) != 6:
                        logger.verbose(
                            'Skipping {}, the URL has an unexpected pattern: {}', episode_id, url
                        )
                        continue  # something is wrong; skip this episode

                    episode_name = list_item.find('h2')
                    if episode_name:
                        title = '{} ({})'.format(next(episode_name.stripped_strings), episode_id)
                    else:
                        title = '{}'.format(episode_id)

                    timer = '0'
                    timerdiv = list_item.find('div', class_='npo-asset-tile-timer')
                    if timerdiv:  # some entries are missing a running time
                        timer = next(timerdiv.stripped_strings)
                    remove_url = list_item.find('div', class_='npo-asset-tile-delete')

                    if premium and not download_premium:
                        logger.debug('Skipping {}, no longer available without premium', title)
                        continue

                    entry_date = url.split('/')[-2]
                    entry_date = self._parse_date(entry_date)

                    if max_age >= 0 and date.today() - entry_date > timedelta(days=max_age):
                        logger.debug('Skipping {}, aired on {}', title, entry_date)
                        continue

                    e = Entry(series_info)
                    e['url'] = url
                    e['title'] = title
                    e['series_name'] = series_info['npo_name']
                    e['series_name_plain'] = self._convert_plain(series_info['npo_name'])
                    e['series_date'] = entry_date
                    e['series_id_type'] = 'date'
                    e['npo_id'] = episode_id
                    e['npo_premium'] = premium
                    e['npo_runtime'] = timer.strip('min').strip()
                    e['language'] = series_info['npo_language']

                    if remove_url and remove_url['data-link']:
                        e['remove_url'] = remove_url['data-link']
                        if config.get('remove_accepted'):
                            e.on_complete(self.entry_complete, task=task)
                    entries.append(e)

        return entries

    def on_task_input(self, task, config):
        # On https://www.npostart.nl/account-profile get the value for the profileId that is in use
        # Details of this account you can get at: https://www.npostart.nl/api/account/@me
        account_profile_url = 'https://www.npostart.nl/account-profile'

        email = config.get('email')
        logger.verbose('Retrieving NPOStart profiles for account {}', email)

        profilejson = self._get_page(task, config, account_profile_url).json()
        if 'profileId' not in profilejson:
            raise plugin.PluginError('Failed to fetch profile for NPO account')
        profileId = profilejson['profileId']
        entries = self._get_favourites_entries(task, config, profileId)
        return entries

    def entry_complete(self, e, task=None):
        if not e.accepted:
            logger.warning('Not removing {} entry {}', e.state, e['title'])
        elif 'remove_url' not in e:
            logger.warning('No remove_url for {}, skipping', e['title'])
        elif task.options.test:
            logger.info('Would remove from watchlist: {}', e['title'])
        else:
            logger.info('Removing from watchlist: {}', e['title'])

            headers = {
                'Origin': 'https://www.npostart.nl',
                'Referer': 'https://www.npostart.nl/mijn_npo',
                'X-XSRF-TOKEN': requests.cookies['XSRF-TOKEN'],
                'X-Requested-With': 'XMLHttpRequest',
            }

            try:
                delete_response = requests.post(
                    e['remove_url'], headers=headers, cookies=requests.cookies
                )
            except HTTPError as error:
                logger.error(
                    'Failed to remove {}, got status {}', e['title'], error.response.status_code
                )
            else:
                if delete_response.status_code != requests.codes.ok:
                    logger.warning(
                        'Failed to remove {}, got status {}',
                        e['title'],
                        delete_response.status_code,
                    )


@event('plugin.register')
def register_plugin():
    plugin.register(NPOWatchlist, 'npo_watchlist', api_ver=2, interfaces=['task'])



# -*- coding: utf-8 -*-
import re
from urllib.parse import quote_plus

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.components.sites.utils import normalize_scene, torrent_availability
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.plugin import PluginError
from flexget.utils.requests import Session as RequestSession
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='fuzer')

requests = RequestSession()

CATEGORIES = {
    # Movies
    'HD Movies': 9,
    'XviD': 7,
    'BRRip': 59,
    'Israeli HD Movies': 61,
    'Israeli Movies': 60,
    'DVDR': 58,
    'Dubbed Movies': 83,
    # TV
    'HD Shows': 10,
    'Shows': 8,
    'Israeli HD Shows': 63,
    'Israeli Shows': 62,
    'Dubbed Shows': 84,
    # Anime
    'Anime': 65,
    # FuzePacks
    'Movie Packs': 73,
    'Shows Packs': 76,
}


class UrlRewriteFuzer:
    schema = {
        'type': 'object',
        'properties': {
            'cookie_password': {'type': 'string'},
            'user_id': {'type': 'integer'},
            'rss_key': {'type': 'string'},
            'category': one_or_more(
                {'oneOf': [{'type': 'string', 'enum': list(CATEGORIES)}, {'type': 'integer'}]}
            ),
        },
        'required': ['user_id', 'cookie_password', 'rss_key'],
        'additionalProperties': False,
    }

    def get_fuzer_soup(self, search_term, categories_list):
        params = {'matchquery': 'any', 'ref_': 'advanced'}
        query = '{}&{}'.format(search_term, '&'.join(categories_list))
        try:
            page = requests.get(
                'https://www.fuzer.me/browse.php?query={}'.format(query),
                params=params,
                cookies=self.cookies,
            )
        except RequestException as e:
            raise PluginError('Could not connect to Fuzer: {}'.format(e))

        if 'login' in page.url:
            raise PluginError('Could not fetch results from Fuzer. Check config')

        logger.debug('Using {} as fuzer search url', page.url)
        return get_soup(page.content)

    def extract_entry_from_soup(self, soup):
        table = soup.find('div', {'id': 'main_table'})
        if table is None:
            raise PluginError('Could not fetch results table from Fuzer, aborting')

        logger.trace('fuzer results table: {}', table)
        table = table.find('table', {'class': 'table_info'})
        if len(table.find_all('tr')) == 1:
            logger.debug('No search results were returned from Fuzer, continuing')
            return []

        entries = []
        for tr in table.find_all("tr"):
            if not tr.get('class') or 'colhead_dark' in tr.get('class'):
                continue
            name = tr.find('div', {'class': 'main_title'}).find('a').text
            torrent_name = re.search(
                '\\n(.*)', tr.find('div', {'style': 'float: right;'}).find('a')['title']
            ).group(1)
            attachment_link = tr.find('div', {'style': 'float: right;'}).find('a')['href']
            attachment_id = re.search(r'attachmentid=(\d+)', attachment_link).group(1)
            raw_size = tr.find_all('td', {'class': 'inline_info'})[0].text.strip()
            seeders = int(tr.find_all('td', {'class': 'inline_info'})[2].text)
            leechers = int(tr.find_all('td', {'class': 'inline_info'})[3].text)

            e = Entry()
            e['title'] = name
            final_url = 'https://www.fuzer.me/rss/torrent.php/{}/{}/{}/{}'.format(
                attachment_id, self.user_id, self.rss_key, torrent_name
            )

            logger.debug('RSS-ified download link: {}', final_url)
            e['url'] = final_url

            e['torrent_seeds'] = seeders
            e['torrent_leeches'] = leechers
            e['torrent_availability'] = torrent_availability(
                e['torrent_seeds'], e['torrent_leeches']
            )

            size = re.search(r'(\d+(?:[.,]\d+)*)\s?([KMGTP]B)', raw_size)
            e['content_size'] = parse_filesize(size.group(0))

            entries.append(e)
        return entries

    @plugin.internet(logger)
    def search(self, task, entry, config=None):
        """
        Search for name from fuzer.
        """
        self.rss_key = config['rss_key']
        self.user_id = config['user_id']

        self.cookies = {
            'fzr2lastactivity': '0',
            'fzr2lastvisit': '',
            'fzr2password': config['cookie_password'],
            'fzr2sessionhash': '',
            'fzr2userid': str(self.user_id),
        }

        category = config.get('category', [0])
        # Make sure categories is a list
        if not isinstance(category, list):
            category = [category]

        # If there are any text categories, turn them into their id number
        categories = [c if isinstance(c, int) else CATEGORIES[c] for c in category]
        c_list = ['c{}={}'.format(quote_plus('[]'), c) for c in categories]

        entries = []
        if entry.get('imdb_id'):
            logger.debug("imdb_id '{}' detected, using in search.", entry['imdb_id'])
            soup = self.get_fuzer_soup(entry['imdb_id'], c_list)
            entries = self.extract_entry_from_soup(soup)
            if entries:
                for e in list(entries):
                    e['imdb_id'] = entry.get('imdb_id')
        else:
            for search_string in entry.get('search_strings', [entry['title']]):
                query = normalize_scene(search_string)
                text = quote_plus(query.encode('windows-1255'))
                soup = self.get_fuzer_soup(text, c_list)
                entries += self.extract_entry_from_soup(soup)
        return (
            sorted(entries, reverse=True, key=lambda x: x.get('torrent_availability'))
            if entries
            else []
        )


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteFuzer, 'fuzer', interfaces=['search'], api_ver=2)



from colorclass.toggles import disable_all_colors
from loguru import logger

from flexget import options
from flexget.event import event
from flexget.plugin import get_plugins
from flexget.terminal import TerminalTable, TerminalTableError, colorize, console, table_parser

logger = logger.bind(name='plugins')


def plugins_summary(manager, options):
    if options.table_type == 'porcelain':
        disable_all_colors()
    header = ['Keyword', 'Interfaces', 'Phases', 'Flags']
    table_data = [header]
    for plugin in sorted(get_plugins(phase=options.phase, interface=options.interface)):
        if options.builtins and not plugin.builtin:
            continue

        flags = []
        if plugin.instance.__doc__:
            flags.append('doc')
        if plugin.builtin:
            flags.append('builtin')
        if plugin.debug:
            if not options.debug:
                continue
            flags.append('developers')

        handlers = plugin.phase_handlers
        roles = []
        for phase in handlers:
            priority = handlers[phase].priority
            roles.append('{0}({1})'.format(phase, priority))

        name = colorize('green', plugin.name) if 'builtin' in flags else plugin.name
        table_data.append([name, ', '.join(plugin.interfaces), ', '.join(roles), ', '.join(flags)])

    try:
        table = TerminalTable(options.table_type, table_data, wrap_columns=[1, 2])
        console(table.output)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))
        return
    console(colorize('green', ' Built-in plugins'))


@event('options.register')
def register_parser_arguments():
    parser = options.register_command(
        'plugins',
        plugins_summary,
        help='Print registered plugin summaries',
        parents=[table_parser],
    )
    parser.add_argument('--interface', help='Show plugins belonging to this interface')
    parser.add_argument('--phase', help='Show plugins that act on this phase')
    parser.add_argument('--builtins', action='store_true', help='Show just builtin plugins')



from collections.abc import MutableSet

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests

logger = logger.bind(name='sonarr_list')

SERIES_ENDPOINT = 'series'
LOOKUP_ENDPOINT = 'series/lookup'
PROFILE_ENDPOINT = 'profile'
ROOTFOLDER_ENDPOINT = 'Rootfolder'
DELETE_ENDPOINT = 'series/{}'

# Sonarr qualities that do no exist in Flexget
QUALITY_MAP = {'Raw-HD': 'remux', 'DVD': 'dvdrip'}


class SonarrSet(MutableSet):
    supported_ids = ['tvdb_id', 'tvrage_id', 'tvmaze_id', 'imdb_id', 'slug', 'sonarr_id']
    schema = {
        'type': 'object',
        'properties': {
            'base_url': {'type': 'string', 'default': 'http://localhost'},
            'base_path': {'type': 'string', 'default': ''},
            'port': {'type': 'number', 'default': 80},
            'api_key': {'type': 'string'},
            'include_ended': {'type': 'boolean', 'default': True},
            'only_monitored': {'type': 'boolean', 'default': True},
            'include_data': {'type': 'boolean', 'default': False},
            'search_missing_episodes': {'type': 'boolean', 'default': True},
            'ignore_episodes_without_files': {'type': 'boolean', 'default': False},
            'ignore_episodes_with_files': {'type': 'boolean', 'default': False},
            'profile_id': {'type': 'integer', 'default': 1},
            'season_folder': {'type': 'boolean', 'default': False},
            'monitored': {'type': 'boolean', 'default': True},
            'root_folder_path': {'type': 'string'},
            'series_type': {
                'type': 'string',
                'enum': ['standard', 'daily', 'anime'],
                'default': 'standard',
            },
            "tags": {"type": "array", "items": {'type': 'string'}},
        },
        'required': ['api_key'],
        'additionalProperties': False,
    }

    def __init__(self, config):
        self.config = config
        self._shows = None

        # cache tags
        self._tags = None

    def _sonarr_request(self, endpoint, term=None, method='get', data=None):
        base_url = self.config['base_url']
        port = self.config['port']
        base_path = self.config['base_path']
        url = '{}:{}{}/api/{}'.format(base_url, port, base_path, endpoint)
        headers = {'X-Api-Key': self.config['api_key']}
        if term:
            url += '?term={}'.format(term)
        try:
            rsp = requests.request(method, url, headers=headers, json=data)
            data = rsp.json()
            logger.trace('sonarr response: {}', data)
            return data
        except RequestException as e:
            base_msg = 'Sonarr returned an error. {}'
            if e.response is not None:
                error = e.response.json()[0]
                error = '{}: {} \'{}\''.format(
                    error['errorMessage'], error['propertyName'], error['attemptedValue']
                )
            else:
                error = str(e)
            raise plugin.PluginError(base_msg.format(error))

    def translate_quality(self, quality_name):
        """
        Translate Sonarr's qualities to ones recognize by Flexget
        """
        if quality_name in QUALITY_MAP:
            return QUALITY_MAP[quality_name]
        return quality_name.replace('-', ' ').lower()

    def quality_requirement_builder(self, quality_profile):
        allowed_qualities = [
            self.translate_quality(quality['quality']['name'])
            for quality in quality_profile['items']
            if quality['allowed']
        ]
        cutoff = self.translate_quality(quality_profile['cutoff']['name'])

        return allowed_qualities, cutoff

    def get_tag_ids(self, entry):
        tags_ids = []

        if not self._tags:
            self._tags = {t["label"].lower(): t["id"] for t in self._sonarr_request("tag")}

        for tag in self.config.get("tags", []):
            if isinstance(tag, int):
                # Handle tags by id
                if tag not in self._tags.values():
                    logger.error(
                        'Unable to add tag with id {} to entry {} as the tag does not exist in sonarr',
                        entry,
                        tag,
                    )
                    continue
                tags_ids.append(tag)
            else:
                tag = entry.render(tag).lower()
                found = self._tags.get(tag)
                if not found:
                    logger.verbose('Adding missing tag {} to Sonarr', tag)
                    found = self._sonarr_request("tag", method="post", data={"label": tag})["id"]
                    self._tags[tag] = found
                tags_ids.append(found)
        return tags_ids

    def list_entries(self, filters=True):
        shows = self._sonarr_request(SERIES_ENDPOINT)
        profiles_dict = {}
        # Retrieves Sonarr's profile list if include_data is set to true
        include_data = self.config.get('include_data')
        if include_data:
            profiles = self._sonarr_request(PROFILE_ENDPOINT)
            profiles_dict = {profile['id']: profile for profile in profiles}

        entries = []
        for show in shows:
            fg_qualities = []  # Initializes the quality parameter
            fg_cutoff = None
            path = show.get('path') if include_data else None
            if filters:
                # Checks if to retrieve just monitored shows
                if not show['monitored'] and self.config.get('only_monitored'):
                    continue
                # Checks if to retrieve ended shows
                if show['status'] == 'ended' and not self.config.get('include_ended'):
                    continue
            profile = profiles_dict.get(show['profileId'])
            if profile:
                fg_qualities, fg_cutoff = self.quality_requirement_builder(profile)

            entry = Entry(
                title=show['title'],
                url='',
                series_name=show['title'],
                tvdb_id=show.get('tvdbId'),
                tvrage_id=show.get('tvRageId'),
                tvmaze_id=show.get('tvMazeId'),
                imdb_id=show.get('imdbid'),
                slug=show.get('titleSlug'),
                sonarr_id=show.get('id'),
            )
            if len(fg_qualities) > 1:
                entry['configure_series_qualities'] = fg_qualities
            elif len(fg_qualities) == 1:
                entry['configure_series_quality'] = fg_qualities[0]
            if path:
                entry['configure_series_path'] = path
            if fg_cutoff:
                entry['configure_series_target'] = fg_cutoff

            if entry.isvalid():
                logger.debug('returning entry {}', entry)
                entries.append(entry)
            else:
                logger.error('Invalid entry created? {}', entry)
                continue

        return entries

    def add_show(self, entry):
        logger.debug('searching for show match for {} using Sonarr', entry)
        term = 'tvdb:{}'.format(entry['tvdb_id']) if entry.get('tvdb_id') else entry['title']
        lookup_results = self._sonarr_request(LOOKUP_ENDPOINT, term=term)
        if not lookup_results:
            logger.debug('could not find series match to {}', entry)
            return
        elif len(lookup_results) > 1:
            logger.debug('got multiple results for Sonarr, using first one')
        show = lookup_results[0]

        if show.get("id"):
            logger.debug('entry {} already exists in Sonarr list as show {}', entry, show)
            return show

        logger.debug('using show {}', show)

        # Getting root folder
        if self.config.get('root_folder_path'):
            root_path = self.config['root_folder_path']
        else:
            root_folder = self._sonarr_request(ROOTFOLDER_ENDPOINT)
            root_path = root_folder[0]['path']

        # Setting defaults for Sonarr
        show['profileId'] = self.config.get('profile_id')
        show['qualityProfileId'] = self.config.get('profile_id')
        show['seasonFolder'] = self.config.get('season_folder')
        show['monitored'] = self.config.get('monitored')
        show['seriesType'] = self.config.get('series_type')
        show['tags'] = self.get_tag_ids(entry)
        show['rootFolderPath'] = root_path
        show['addOptions'] = {
            "ignoreEpisodesWithFiles": self.config.get('ignore_episodes_with_files'),
            "ignoreEpisodesWithoutFiles": self.config.get('ignore_episodes_without_files'),
            "searchForMissingEpisodes": self.config.get('search_missing_episodes'),
        }

        logger.debug('adding show {} to sonarr', show)
        returned_show = self._sonarr_request(SERIES_ENDPOINT, method='post', data=show)
        return returned_show

    def remove_show(self, show):
        logger.debug('sending sonarr delete show request')
        self._sonarr_request(DELETE_ENDPOINT.format(show['sonarr_id']), method='delete')

    def shows(self, filters=True):
        if self._shows is None:
            self._shows = self.list_entries(filters=filters)
        return self._shows

    def _find_entry(self, entry, filters=True):
        for show in self.shows(filters=filters):
            if any(
                entry.get(id) is not None and entry[id] == show[id] for id in self.supported_ids
            ):
                return show
            if entry.get('title').lower() == show.get('title').lower():
                return show

    def _from_iterable(self, it):
        # TODO: is this the right answer? the returned object won't have our custom __contains__ logic
        return set(it)

    def __iter__(self):
        return (entry for entry in self.shows())

    def __len__(self):
        return len(self.shows())

    def __contains__(self, entry):
        return self._find_entry(entry) is not None

    def add(self, entry):
        if not self._find_entry(entry, filters=False):
            show = self.add_show(entry)
            if show:
                self._shows = None
                logger.verbose('Successfully added show {} to Sonarr', show['title'])
        else:
            logger.debug('entry {} already exists in Sonarr list', entry)

    def discard(self, entry):
        show = self._find_entry(entry, filters=False)
        if not show:
            logger.debug('Did not find matching show in Sonarr for {}, skipping', entry)
            return
        self.remove_show(show)
        logger.verbose('removed show {} from Sonarr', show['title'])

    @property
    def immutable(self):
        return False

    @property
    def online(self):
        """Set the online status of the plugin, online plugin should be treated differently in certain situations,
        like test mode"""
        return True

    def get(self, entry):
        return self._find_entry(entry)


class SonarrList:
    schema = SonarrSet.schema

    @staticmethod
    def get_list(config):
        return SonarrSet(config)

    def on_task_input(self, task, config):
        return list(SonarrSet(config))


@event('plugin.register')
def register_plugin():
    plugin.register(SonarrList, 'sonarr_list', api_ver=2, interfaces=['task', 'list'])



import re
from urllib.parse import parse_qs, urlparse

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.event import event
from flexget.utils.requests import Session, TimedLimiter
from flexget.utils.soup import get_soup

logger = logger.bind(name='google')

requests = Session()
requests.headers.update({'User-Agent': 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)'})
requests.add_domain_limiter(TimedLimiter('imdb.com', '2 seconds'))


class UrlRewriteGoogleCse:
    """Google custom query urlrewriter."""

    # urlrewriter API
    def url_rewritable(self, task, entry):
        if entry['url'].startswith('http://www.google.com/cse?'):
            return True
        if entry['url'].startswith('http://www.google.com/custom?'):
            return True
        return False

    # urlrewriter API
    def url_rewrite(self, task, entry):
        try:
            # need to fake user agent
            txheaders = {'User-agent': 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)'}
            page = task.requests.get(entry['url'], headers=txheaders)
            soup = get_soup(page.text)
            results = soup.find_all('a', attrs={'class': 'l'})
            if not results:
                raise UrlRewritingError('No results')
            for res in results:
                url = res.get('href')
                url = url.replace('/interstitial?url=', '')
                # generate match regexp from google search result title
                regexp = '.*'.join([x.contents[0] for x in res.find_all('em')])
                if re.match(regexp, entry['title']):
                    logger.debug('resolved, found with {}', regexp)
                    entry['url'] = url
                    return
            raise UrlRewritingError('Unable to resolve')
        except Exception as e:
            raise UrlRewritingError(e)


class UrlRewriteGoogle:
    # urlrewriter API

    def url_rewritable(self, task, entry):
        if entry['url'].startswith('https://www.google.com/search?q='):
            return True
        return False

    # urlrewriter API
    def url_rewrite(self, task, entry):
        logger.debug('Requesting {}', entry['url'])
        page = requests.get(entry['url'])
        soup = get_soup(page.text)

        for link in soup.findAll('a', attrs={'href': re.compile(r'^/url')}):
            # Extract correct url from google internal link
            href = 'http://google.com' + link['href']
            args = parse_qs(urlparse(href).query)
            href = args['q'][0]

            # import IPython; IPython.embed()
            # import sys
            # sys.exit(1)
            # href = link['href'].lstrip('/url?q=').split('&')[0]

            # Test if entry with this url would be recognized by some urlrewriter
            logger.trace('Checking if {} is known by some rewriter', href)
            fake_entry = {'title': entry['title'], 'url': href}
            urlrewriting = plugin.get('urlrewriting', self)
            if urlrewriting.url_rewritable(task, fake_entry):
                logger.debug('--> rewriting {} (known url pattern)', href)
                entry['url'] = href
                return
            else:
                logger.debug('<-- ignoring {} (unknown url pattern)', href)
        raise UrlRewritingError('Unable to resolve')


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteGoogleCse, 'google_cse', interfaces=['urlrewriter'], api_ver=2)
    plugin.register(UrlRewriteGoogle, 'google', interfaces=['urlrewriter'], api_ver=2)
